import ssl
import logging
import traceback
from typing import Set, Dict, List, Tuple, Optional

from qm.utils import SERVICE_HEADER_NAME
from qm.utils.async_utils import run_async
from qm.api.frontend_api import FrontendApi
from qm.api.models.info import QuaMachineInfo
from qm.api.models.debug_data import DebugData
from qm.exceptions import QmServerDetectionError
from qm.api.info_service_api import InfoServiceApi
from qm.communication.http_redirection import send_redirection_check
from qm.api.models.server_details import BASE_TIMEOUT, MAX_MESSAGE_SIZE, ServerDetails, ConnectionDetails

logger = logging.getLogger(__name__)

DEFAULT_PORTS = (80,)


def detect_server(
    cluster_name: Optional[str],
    user_token: str,
    ssl_context: Optional[ssl.SSLContext],
    host: str,
    port_from_user_config: Optional[int],
    user_provided_port: Optional[int],
    add_debug_data: bool,
    timeout: Optional[float] = None,
    max_message_size: Optional[int] = None,
    extra_headers: Optional[Dict[str, str]] = None,
) -> ServerDetails:
    ports_to_try = _get_ports(port_from_user_config, user_provided_port)
    extra_headers = extra_headers or {}
    headers = _create_headers(extra_headers, "gateway", cluster_name, user_token)
    errors: List[Tuple[str, str]] = []

    for port in ports_to_try:
        dst = f"{host}:{port}"
        logger.debug(f"Probing gateway at: {dst}")
        debug_data = DebugData()

        connection_details = ConnectionDetails(
            host=host,
            port=port,
            user_token=user_token,
            ssl_context=ssl_context,
            max_message_size=max_message_size if max_message_size else MAX_MESSAGE_SIZE,
            headers=headers,
            timeout=timeout if timeout else BASE_TIMEOUT,
            debug_data=debug_data if add_debug_data else None,
        )

        try:
            connection_details, octaves = _redirect(connection_details)
            info, server_version = _connect(connection_details)
        except Exception as e:
            errors.append((dst, f"{e}\n{traceback.format_exc()}"))
            continue

        if not server_version:
            errors.append((dst, "could not get server version"))
            continue

        logger.debug(f"Gateway discovered at: {dst}")
        return ServerDetails(
            port=port,
            host=host,
            server_version=server_version,
            qua_implementation=info,
            connection_details=connection_details,
            octaves={
                name: ConnectionDetails(
                    octave_host, octave_port, headers=_create_headers(extra_headers, name, cluster_name, user_token)
                )
                for name, (octave_host, octave_port) in octaves.items()
            },
        )

    targets = ",".join([f"{host}:{port}" for port in ports_to_try])
    cluster_str = f"cluster '{cluster_name}'" if cluster_name else "a cluster"
    message = (
        f"Failed to detect to QuantumMachines server, failed to connect to {cluster_str}. "
        f"Tried connecting to {targets}."
    )
    errors_msgs = "\n".join([f"{dst}: {error}" for dst, error in errors])
    logger.error(f"{message}\nErrors:\n{errors_msgs}.")
    raise QmServerDetectionError(message)


def _redirect(connection_details: ConnectionDetails) -> Tuple[ConnectionDetails, Dict[str, Tuple[str, int]]]:
    response_details = run_async(
        send_redirection_check(
            connection_details.host, connection_details.port, connection_details.headers, connection_details.timeout
        )
    )
    octaves = {}
    if response_details.host != connection_details.host or response_details.port != connection_details.port:
        logger.debug(
            f"Connection redirected from '{connection_details.host}:{connection_details.port}' to "
            f"'{response_details.host}:{response_details.port}'"
        )
        connection_details.host = response_details.host
        connection_details.port = response_details.port
        octaves = response_details.octaves
        if octaves:
            logger.debug(f"Detected octaves: {octaves}")

    return connection_details, octaves


def _connect(
    connection_details: ConnectionDetails,
) -> Tuple[Optional[QuaMachineInfo], Optional[str]]:
    frontend = FrontendApi(connection_details)
    info_service = InfoServiceApi(connection_details)

    info = info_service.get_info()
    if info.implementation.version:
        server_version = info.implementation.version
    else:
        server_version = frontend.get_version()

    logger.debug(f"Established connection to {connection_details.host}:{connection_details.port}")

    return info, server_version


def _get_ports(port_from_config: Optional[int], user_provided_port: Optional[int]) -> Set[int]:
    if user_provided_port is not None:
        return {user_provided_port}

    ports = set()
    if port_from_config is not None:
        ports.add(int(port_from_config))

    ports.update({int(port) for port in DEFAULT_PORTS})
    return ports


def _create_headers(
    base_headers: Dict[str, str], device_identity: str, cluster_name: Optional[str], user_token: Optional[str]
) -> Dict[str, str]:
    headers = {}
    headers.update(base_headers if base_headers is not None else {})
    headers[SERVICE_HEADER_NAME] = device_identity
    if user_token:
        headers["authorization"] = f"Bearer {user_token}"
    headers.update(_create_cluster_headers(cluster_name))
    return headers


def _create_cluster_headers(cluster_name: Optional[str]) -> Dict[str, str]:
    if cluster_name and cluster_name != "any":
        return {"cluster_name": cluster_name}
    return {"cluster_name": "any", "any_cluster": "true"}



import logging
import warnings
from collections import defaultdict
from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Any, Dict, List, Tuple, Union, Generic, TypeVar, Optional, ContextManager

from octave_sdk.octave import ClockInfo
from octave_sdk import IFMode, ClockType, RFOutputMode, ClockFrequency, OctaveLOSource, RFInputLOSource

from qm.elements.element import Element
from qm.grpc.qua_config import QuaConfigMixInputs
from qm.elements.element_outputs import DownconvertedOutput
from qm.elements.up_converted_input import UpconvertedInput
from qm.octave.octave_manager import ClockMode, OctaveManager
from qm.octave.abstract_calibration_db import AbstractIFCalibration, AbstractLOCalibration
from qm.type_hinting.config_types import DictQuaConfig, MixerConfigType, OPX1000ControllerConfigType
from qm.octave.octave_mixer_calibration import DeprecatedCalibrationResult, convert_to_old_calibration_result

if TYPE_CHECKING:
    from qm.api.v2.qm_api import QmApi
    from qm.api.v2.job_api import JobApi
    from qm.quantum_machine import QuantumMachine
    from qm.jobs.running_qm_job import RunningQmJob

logger = logging.getLogger(__name__)


class ElementHasNoOctaveError(Exception):
    def __init__(self, name: str):
        self._name = name

    def __str__(self) -> str:
        return f"Element {self._name} has no octave connected to it."


class NoCalibrationDbAttachedError(Exception):
    def __init__(self, name: str):
        self._name = name

    def __str__(self) -> str:
        return f"No calibration database attached to element {self._name}"


class NoCalibrationParamsError(Exception):
    def __init__(self, name: str):
        self._name = name

    def __str__(self) -> str:
        return f"No calibration parameters found for element {self._name}"


QmInstT = TypeVar("QmInstT", "QuantumMachine", "QmApi")
JobInstT = TypeVar("JobInstT", "RunningQmJob", "JobApi")


class QmOctaveBase(Generic[QmInstT, JobInstT], metaclass=ABCMeta):
    def __init__(self, qm: QmInstT, octave_manager: OctaveManager):
        self._qm: QmInstT = qm
        self._octave_manager = octave_manager

    def _get_upconverted_input(self, name: str) -> UpconvertedInput:
        instance = self._qm._elements[name]
        if not isinstance(instance.input, UpconvertedInput):
            raise ElementHasNoOctaveError(name)
        return instance.input

    def _get_downconverted_output(self, name: str) -> DownconvertedOutput:
        instance = self._qm._elements[name]
        if not isinstance(instance.output, DownconvertedOutput):
            raise ElementHasNoOctaveError(name)
        return instance.output

    def start_batch_mode(self) -> None:
        self._octave_manager.start_batch_mode()

    def end_batch_mode(self) -> None:
        self._octave_manager.end_batch_mode()

    def batch_mode(self) -> ContextManager[None]:
        return self._octave_manager.batch_mode()

    def set_qua_element_octave_rf_in_port(self, element: str, octave_name: str, rf_input_index: int) -> None:
        """Sets the octave down conversion port for the given element.

        Args:
            element (str): The name of the element
            octave_name (str): The octave name
            rf_input_index (RFInputRFSource): input index - can be 1 or 2
        """
        inst = self._qm._elements[element]
        client = self._octave_manager.get_client(octave_name)
        rf_input_client = client.rf_inputs[rf_input_index]
        inst.output = DownconvertedOutput(rf_input_client)

    def load_lo_frequency_from_config(self, elements: Union[List[str], str]) -> None:
        """Loads into the octave synthesizers the LO frequencies specified for elements
        in the element list

        Args:
            elements: A list of elements to load LO frequencies for from the configuration
        """
        warnings.warn(
            "lo_frequency is now set directly from config when a QuantumMachine is created, no need to do it directly. "
            "If you want, you can run over the elements and do set_frequency.",
            category=DeprecationWarning,
            stacklevel=2,
        )

    def set_lo_frequency(self, element: str, lo_frequency: float, set_source: bool = True) -> None:
        """Sets the LO frequency of the synthesizer associated with the given element. Will not change the synthesizer if set_source = False

        Args:
            element (str): The name of the element
            lo_frequency (float): The LO frequency
            set_source (Boolean): Set the synthesizer (True) or just
                update the local database (False)
        """
        input_inst = self._get_upconverted_input(element)
        input_inst.set_lo_frequency(lo_frequency, set_source)

    def update_external_lo_frequency(
        self,
        element: str,
        lo_frequency: float,
    ) -> None:
        """Updates the local database on the external LO frequency (provided by the user)
        associated with element

        Args:
            element (str): The name of the element
            lo_frequency (float): The LO frequency
        """
        input_inst = self._get_upconverted_input(element)
        input_inst.inform_element_about_lo_frequency(lo_frequency)

    def set_lo_source(self, element: str, lo_port: OctaveLOSource) -> None:
        """Associate the given LO source with the given element. Always be sure the given LO source is internally connected to the element according to the Octave Block Diagram in the documentation

        Args:
            element (str): The name of the element
            lo_port (OctaveLOSource): One of the allowed sources
                according the internal connectivity
        """
        input_inst = self._get_upconverted_input(element)
        input_inst.set_lo_source(lo_port)

    def set_rf_output_mode(self, element: str, switch_mode: RFOutputMode) -> None:
        """Configures the output switch of the upconverter associated to element.
        switch_mode can be either: 'on', 'off', 'trig_normal' or 'trig_inverse':

        - 'trig_normal' mode a high trigger will turn the switch on and a low trigger will turn it off
        - 'trig_inverse' mode a high trigger will turn the switch off and a low trigger will turn it on
        - 'on' the switch will be permanently on.
        - 'off' mode the switch will be permanently off.

        Args:
            element (str): The name of the element
            switch_mode (RFOutputMode): switch mode according to the
                allowed states
        """
        input_inst = self._get_upconverted_input(element)
        input_inst.set_rf_output_mode(switch_mode)

    def set_rf_output_gain(self, element: str, gain_in_db: float) -> None:
        """Sets the RF output gain for the up-converter associated with the element.
        RF_gain is in steps of 0.5dB from -20 to +20 and is referring
        to the maximum OPX output power of 4dBm (=0.5V pk-pk) So for a value of -20
        for example, an IF signal coming from the OPX at max power (4dBm) will be
        upconverted and come out of Octave at -16dBm

        Args:
            element (str): The name of the element
            gain_in_db (float): The RF output gain in dB
        """
        # a. use value custom set in qm.octave.update_external
        # b. use value from config
        input_inst = self._get_upconverted_input(element)
        input_inst.set_rf_output_gain(gain_in_db)

    def set_downconversion(
        self,
        element: str,
        lo_source: Optional[RFInputLOSource] = None,
        lo_frequency: Optional[float] = None,
        if_mode_i: IFMode = IFMode.direct,
        if_mode_q: IFMode = IFMode.direct,
        disable_warning: bool = False,
    ) -> None:
        """Sets the LO source and frequency for the downconverter of a given element. Sets also the mode of the I and Q lines after downconversion

        - The LO source will be the one associated with the element's upconversion.
        - If only the element is given, the LO source and frequency for downconversion will be the same as those set for the upconversion of the element.
        - IFMode sets the I/Q lines mode after downconversion and can be:
            - direct: bypass
            - envelope: goes through an envelope detector for advanced applications
            - Mixer: goes through a low frequency mixer for advanced applications

        Args:
            element (str): The name of the element
            lo_source (RFInputLOSource): LO source according to the allowed LO sources
            lo_frequency (float): The LO frequency
            if_mode_i (IFMode): Sets the I channel mode of the downconverter
            if_mode_q (IFMode): Sets the Q channel mode of the downconverter
            disable_warning (Boolean): Disable warnings about non-matching LO sources and elements if True
        """
        if lo_source is None:
            up_conv = self._get_upconverted_input(element)
            lo_source = up_conv.lo_source
        inst = self._get_downconverted_output(element)
        inst.set_downconversion(lo_source, lo_frequency, if_mode_i, if_mode_q)

    def set_use_input_attenuators(self, element: str, use_input_attenuators: bool) -> None:
        """Sets the use of IQ attenuators for the downconverter associated with the element.

        Args:
            element (str): The name of the element
            use_input_attenuators (bool): True to use IQ attenuators, False otherwise
        """
        input_inst = self._get_upconverted_input(element)
        input_inst.set_use_input_attenuators(use_input_attenuators)

    def calibrate_element(
        self,
        element: str,
        lo_if_frequencies_tuple_list: Optional[List[Tuple[int, int]]] = None,
        save_to_db: bool = True,
        offset_frequency: float = 7e6,
        **kwargs: Any,
    ) -> Dict[Tuple[int, int], DeprecatedCalibrationResult]:
        """Calibrate the up converter associated with an element for the given LO & IF frequencies.

        - Frequencies are given as a list of tuples : [(LO,IF1)(LO,IF2)...]
        - The function need to be run for each LO frequency and element separately

        - If close_open_quantum_machines is set to True one must open a new quantum machine after calibration ends
        Args:
            element (str): The name of the element for calibration
            lo_if_frequencies_tuple_list (list): a list of tuples that
                consists of all the (LO,IF) pairs for calibration
            save_to_db (boolean): If true (default), The calibration
                parameters will be saved to the calibration database
        """
        warnings.warn(
            "Calibrate element was moved to the QuantumMachine instance, please use it from there", DeprecationWarning
        )

        if lo_if_frequencies_tuple_list is not None:
            lo_if_dict_tmp: Dict[float, Tuple[float, ...]] = defaultdict(tuple)
            for _lo, _if in lo_if_frequencies_tuple_list:
                lo_if_dict_tmp[float(_lo)] += (_if,)
            lo_if_dict = dict(lo_if_dict_tmp)
        else:
            lo_if_dict = None

        calibration_result = self._qm.calibrate_element(element, lo_if_dict, save_to_db)

        return convert_to_old_calibration_result(calibration_result, self._get_upconverted_input(element).mixer)

    def set_clock(
        self,
        octave_name: str,
        clock_type: Optional[ClockType] = None,
        frequency: Optional[ClockFrequency] = None,
        clock_mode: Optional[ClockMode] = None,
    ) -> None:
        """This function sets the octave clock type - internal, external or buffered.
        It can also set the clock frequency - 10, 100 or 1000 MHz

        - Internal can only be 10 MHz
        - External can be 10, 100, or 1000 MHz
        - Buffered and External behave the same when using 1000 MHz clock
        - ClockType & ClockFrequency will be deprecated soon, users should use clock_mode: ClockMode instead

        Args:
            octave_name (str): The octave name to set clock for
            clock_type (ClockType): clock type according to ClockType
            frequency (ClockFrequency): Clock frequency according to ClockFrequency
            clock_mode (ClockMode): Clock mode according to ClockMode
        """
        self._octave_manager.set_clock(octave_name, clock_type, frequency, clock_mode)

    def get_clock(self, octave_name: str) -> ClockInfo:
        """Gets the clock info for a given octave name

        Args:
            octave_name (str): The octave name to get clock for
        :returns ClockInfo: Info about the clock as an object
        """
        return self._octave_manager.get_clock(octave_name)

    @staticmethod
    def _fetch_calibrations_from_db(
        qe: Element[QuaConfigMixInputs],
    ) -> Tuple[AbstractLOCalibration, AbstractIFCalibration]:
        assert isinstance(qe.input, UpconvertedInput)
        if qe.input._calibration_db is None:
            logger.warning(f"No calibration DB is attached for element {qe.name}, not changing anything.")
            raise NoCalibrationDbAttachedError(qe.name)
        db = qe.input._calibration_db

        lo_cal = db.get_lo_cal(qe.input.port, lo_freq=qe.input.lo_frequency, gain=qe.input.gain)
        if_cal = db.get_if_cal(
            qe.input.port, lo_freq=qe.input.lo_frequency, gain=qe.input.gain, if_freq=qe.intermediate_frequency
        )
        if if_cal is None or lo_cal is None:
            raise NoCalibrationParamsError(qe.name)
        return lo_cal, if_cal

    def set_element_parameters_from_calibration_db(self, element: str, running_job: Optional[JobInstT] = None) -> None:
        """Apply correction parameters to an element from the calibration database. The parameters will be selected
        according to the current LO and IF frequencies in the configuration. Therefore, If used when sweeping the LO,
        it is required to first set the new LO freq using qm.octave.set_lo_frequency() before updating the calibration
        parameters.

        Args:
            element (str): The name of the element for setting the calibration parameters
            running_job: Optional. If given, will update the element's parameters of the given running job
        """
        qe = self._qm._elements[element]
        assert isinstance(qe.input, UpconvertedInput)
        try:
            lo_cal, if_cal = self._fetch_calibrations_from_db(qe)
        except (NoCalibrationParamsError, NoCalibrationDbAttachedError) as err:
            logger.warning(f"Cannot change element params due to the following error: {err}")
            return

        self._update_element_parameters(qe, lo_cal, if_cal)

        if running_job:
            self._set_running_job_parameters(running_job, element, lo_cal, if_cal)

    @abstractmethod
    def _update_element_parameters(
        self, qe: Element[QuaConfigMixInputs], lo_cal: AbstractLOCalibration, if_cal: AbstractIFCalibration
    ) -> None:
        pass

    @abstractmethod
    def _set_running_job_parameters(
        self,
        job: JobInstT,
        element: str,
        lo_cal: AbstractLOCalibration,
        if_cal: AbstractIFCalibration,
    ) -> None:
        pass


class QmOctave(QmOctaveBase["QuantumMachine", "RunningQmJob"]):
    def _update_element_parameters(
        self, qe: Element[QuaConfigMixInputs], lo_cal: AbstractLOCalibration, if_cal: AbstractIFCalibration
    ) -> None:
        assert isinstance(qe.input, UpconvertedInput)
        qe.input.set_output_dc_offset(i_offset=lo_cal.get_i0(), q_offset=lo_cal.get_q0())
        qe.input.set_mixer_correction(
            intermediate_frequency=qe.intermediate_frequency,
            lo_frequency=qe.input.lo_frequency,
            values=if_cal.get_correction(),
        )

    def _set_running_job_parameters(
        self, job: "RunningQmJob", element: str, lo_cal: AbstractLOCalibration, if_cal: AbstractIFCalibration
    ) -> None:
        job.set_element_correction(element, if_cal.get_correction())


class QmOctaveForNewApi(QmOctaveBase["QmApi", "JobApi"]):
    def _update_element_parameters(
        self, qe: Element[QuaConfigMixInputs], lo_cal: AbstractLOCalibration, if_cal: AbstractIFCalibration
    ) -> None:
        assert isinstance(qe.input, UpconvertedInput)
        update = create_dc_offset_octave_update(qe.input, i_offset=lo_cal.get_i0(), q_offset=lo_cal.get_q0())
        update["mixers"] = {
            qe.input.mixer: [
                create_mixer_correction(qe.intermediate_frequency, qe.input.lo_frequency, if_cal.get_correction())
            ]
        }
        self._qm.update_config(update)

    def _set_running_job_parameters(
        self, job: "JobApi", element: str, lo_cal: AbstractLOCalibration, if_cal: AbstractIFCalibration
    ) -> None:
        job.set_output_dc_offset_by_element(element, ("I", "Q"), (lo_cal.get_i0(), lo_cal.get_q0()))
        job.set_element_correction(element, if_cal.get_correction())


def create_dc_offset_octave_update(qe_input: UpconvertedInput, i_offset: float, q_offset: float) -> DictQuaConfig:
    con_i, fem_i, port_i = qe_input.i_port.controller, qe_input.i_port.fem, qe_input.i_port.number
    con_q, fem_q, port_q = qe_input.q_port.controller, qe_input.q_port.fem, qe_input.q_port.number
    controllers: Dict[str, OPX1000ControllerConfigType] = {con_i: {"fems": {}}, con_q: {"fems": {}}}
    controllers[con_i]["fems"][fem_i] = {"analog_outputs": {}, "type": "LF"}  # type: ignore[index]
    controllers[con_q]["fems"][fem_q] = {"analog_outputs": {}, "type": "LF"}  # type: ignore[index]
    controllers[con_i]["fems"][fem_i]["analog_outputs"][port_i] = {"offset": i_offset}  # type: ignore[index]
    controllers[con_q]["fems"][fem_q]["analog_outputs"][port_q] = {"offset": q_offset}  # type: ignore[index]
    update: DictQuaConfig = {"version": 1, "controllers": controllers}
    return update


def create_mixer_correction(
    intermediate_frequency: float, lo_frequency: float, correction: Tuple[float, float, float, float]
) -> MixerConfigType:
    return {
        "correction": correction,
        "intermediate_frequency": intermediate_frequency,
        "lo_frequency": lo_frequency,
    }



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: qm/pb/errors.proto
# plugin: python-betterproto
# This file has been @generated

from dataclasses import dataclass

import betterproto


class JobManagerErrorTypes(betterproto.Enum):
    JobManagerUnspecifiedError = 0
    MissingJobError = 1000
    InvalidJobExecutionStatusError = 1001
    InvalidOperationOnSimulatorJobError = 1002
    InvalidOperationOnRealJobError = 1003
    JobOperationSpecificError = 1004
    ConfigQueryError = 1005
    UnknownInputStreamError = 1006


class JobOperationSpecificErrorTypes(betterproto.Enum):
    JobOperationUnspecifiedError = 0
    SingleInputElementError = 3000
    InvalidCorrectionMatrixError = 3001
    ElementWithoutIntermediateFrequencyError = 3002
    InvalidDigitalInputThresholdError = 3003
    InvalidDigitalInputDeadtimeError = 3004
    InvalidDigitalInputPolarityError = 3005


class ConfigQueryErrorTypes(betterproto.Enum):
    ConfigQueryUnspecifiedError = 0
    MissingControllerError = 4000
    MissingElementError = 4001
    MissingDigitalInputError = 4002



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: qm/pb/inc_qua.proto
# plugin: python-betterproto
# This file has been @generated
import warnings
from dataclasses import dataclass
from typing import (
    List,
    Optional,
)

import betterproto
import betterproto.lib.google.protobuf as betterproto_lib_google_protobuf

from .. import qua_config as _qua_config__


class QuaProgramType(betterproto.Enum):
    INT = 0
    BOOL = 1
    REAL = 2


class QuaProgramChirpUnits(betterproto.Enum):
    HzPerNanoSec = 0
    mHzPerNanoSec = 1
    uHzPerNanoSec = 2
    nHzPerNanoSec = 3
    pHzPerNanoSec = 4


class QuaProgramUpdateFrequencyStatementUnits(betterproto.Enum):
    Hz = 0
    mHz = 1
    uHz = 2
    nHz = 3
    pHz = 4


class QuaProgramUpdateOscillatorFrequencyStatementUnits(betterproto.Enum):
    GHz = 0
    MHz = 1
    KHz = 2
    Hz = 3


class QuaProgramUpdateOscillatorFrequencyStatementUpdateComponentSelection(
    betterproto.Enum
):
    both = 0
    upconverter = 1
    downconverter = 2


class QuaProgramBinaryExpressionBinaryOperator(betterproto.Enum):
    ADD = 0
    SUB = 1
    MULT = 2
    DIV = 3
    AND = 4
    """binary operators"""

    OR = 5
    XOR = 6
    LT = 7
    """comparisons"""

    LET = 8
    GT = 9
    GET = 10
    EQ = 11
    SHL = 12
    """shifts"""

    SHR = 13


@dataclass(eq=False, repr=False)
class QuaProgram(betterproto.Message):
    config: "_qua_config__.QuaConfig" = betterproto.message_field(1)
    dyn_config: "QuaProgramDynamicConfig" = betterproto.message_field(2)
    script: "QuaProgramScript" = betterproto.message_field(5)
    compiler_options: "QuaProgramCompilerOptions" = betterproto.message_field(6)
    result_analysis: "QuaResultAnalysis" = betterproto.message_field(7)
    config_update: "_qua_config__.QuaConfig" = betterproto.message_field(8)


@dataclass(eq=False, repr=False)
class QuaProgramCompilerOptions(betterproto.Message):
    use_experimental_calculation_compiler: bool = betterproto.bool_field(1)
    optimize_merge_code_execution: bool = betterproto.bool_field(2)
    optimize_write_read_commands: bool = betterproto.bool_field(3)
    skip_optimizations: List[str] = betterproto.string_field(4)
    strict: bool = betterproto.bool_field(5)
    flags: List[str] = betterproto.string_field(6)

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.is_set("use_experimental_calculation_compiler"):
            warnings.warn(
                "QuaProgramCompilerOptions.use_experimental_calculation_compiler is deprecated",
                DeprecationWarning,
            )
        if self.is_set("optimize_merge_code_execution"):
            warnings.warn(
                "QuaProgramCompilerOptions.optimize_merge_code_execution is deprecated",
                DeprecationWarning,
            )
        if self.is_set("optimize_write_read_commands"):
            warnings.warn(
                "QuaProgramCompilerOptions.optimize_write_read_commands is deprecated",
                DeprecationWarning,
            )
        if self.is_set("skip_optimizations"):
            warnings.warn(
                "QuaProgramCompilerOptions.skip_optimizations is deprecated",
                DeprecationWarning,
            )
        if self.is_set("strict"):
            warnings.warn(
                "QuaProgramCompilerOptions.strict is deprecated", DeprecationWarning
            )


@dataclass(eq=False, repr=False)
class QuaProgramScript(betterproto.Message):
    variables: List["QuaProgramVarDeclaration"] = betterproto.message_field(1)
    body: "QuaProgramStatementsCollection" = betterproto.message_field(10)


@dataclass(eq=False, repr=False)
class QuaProgramVarDeclaration(betterproto.Message):
    name: str = betterproto.string_field(1)
    type: "QuaProgramType" = betterproto.enum_field(2)
    size: int = betterproto.int32_field(3)
    value: List["QuaProgramLiteralExpression"] = betterproto.message_field(4)
    dim: int = betterproto.int32_field(5)
    is_input_stream: bool = betterproto.bool_field(6)


@dataclass(eq=False, repr=False)
class QuaProgramQuantumElementReference(betterproto.Message):
    loc: str = betterproto.string_field(100)
    name: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class QuaProgramPulseReference(betterproto.Message):
    loc: str = betterproto.string_field(100)
    name: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class QuaProgramIntegrationWeightReference(betterproto.Message):
    loc: str = betterproto.string_field(100)
    name: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class QuaProgramAnyStatement(betterproto.Message):
    play: "QuaProgramPlayStatement" = betterproto.message_field(
        1, group="statement_oneof"
    )
    measure: "QuaProgramMeasureStatement" = betterproto.message_field(
        2, group="statement_oneof"
    )
    wait: "QuaProgramWaitStatement" = betterproto.message_field(
        3, group="statement_oneof"
    )
    sync: "QuaProgramSyncStatement" = betterproto.message_field(
        4, group="statement_oneof"
    )
    if_: "QuaProgramIfStatement" = betterproto.message_field(5, group="statement_oneof")
    assign: "QuaProgramAssignmentStatement" = betterproto.message_field(
        6, group="statement_oneof"
    )
    for_: "QuaProgramForStatement" = betterproto.message_field(
        7, group="statement_oneof"
    )
    rus: "QuaProgramRusStatement" = betterproto.message_field(
        8, group="statement_oneof"
    )
    align: "QuaProgramAlignStatement" = betterproto.message_field(
        9, group="statement_oneof"
    )
    update_frequency: "QuaProgramUpdateFrequencyStatement" = betterproto.message_field(
        10, group="statement_oneof"
    )
    z_rotation: "QuaProgramZRotationStatement" = betterproto.message_field(
        11, group="statement_oneof"
    )
    pause: "QuaProgramPauseStatement" = betterproto.message_field(
        12, group="statement_oneof"
    )
    save: "QuaProgramSaveStatement" = betterproto.message_field(
        13, group="statement_oneof"
    )
    for_each: "QuaProgramForEachStatement" = betterproto.message_field(
        14, group="statement_oneof"
    )
    wait_for_trigger: "QuaProgramWaitForTriggerStatement" = betterproto.message_field(
        15, group="statement_oneof"
    )
    qrun: "QuaProgramQRunStatement" = betterproto.message_field(
        16, group="statement_oneof"
    )
    update_correction: "QuaProgramUpdateCorrectionStatement" = (
        betterproto.message_field(17, group="statement_oneof")
    )
    reset_phase: "QuaProgramResetPhaseStatement" = betterproto.message_field(
        18, group="statement_oneof"
    )
    ramp_to_zero: "QuaProgramRampToZeroStatement" = betterproto.message_field(
        19, group="statement_oneof"
    )
    reset_frame: "QuaProgramResetFrameStatement" = betterproto.message_field(
        20, group="statement_oneof"
    )
    set_dc_offset: "QuaProgramSetDcOffsetStatement" = betterproto.message_field(
        21, group="statement_oneof"
    )
    advance_input_stream: "QuaProgramAdvanceInputStreamStatement" = (
        betterproto.message_field(22, group="statement_oneof")
    )
    strict_timing: "QuaProgramStrictTimingStatement" = betterproto.message_field(
        23, group="statement_oneof"
    )
    fast_frame_rotation: "QuaProgramFastFrameRotationStatement" = (
        betterproto.message_field(24, group="statement_oneof")
    )
    update_oscillator_frequency: "QuaProgramUpdateOscillatorFrequencyStatement" = (
        betterproto.message_field(25, group="statement_oneof")
    )
    reset_global_phase: "QuaProgramResetGlobalPhaseStatement" = (
        betterproto.message_field(27, group="statement_oneof")
    )

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.is_set("qrun"):
            warnings.warn(
                "QuaProgramAnyStatement.qrun is deprecated", DeprecationWarning
            )


@dataclass(eq=False, repr=False)
class QuaProgramAmpMultiplier(betterproto.Message):
    loc: str = betterproto.string_field(100)
    v0: "QuaProgramAnyScalarExpression" = betterproto.message_field(1)
    v1: "QuaProgramAnyScalarExpression" = betterproto.message_field(2)
    v2: "QuaProgramAnyScalarExpression" = betterproto.message_field(3)
    v3: "QuaProgramAnyScalarExpression" = betterproto.message_field(4)


@dataclass(eq=False, repr=False)
class QuaProgramRampToZeroStatement(betterproto.Message):
    loc: str = betterproto.string_field(100)
    qe: "QuaProgramQuantumElementReference" = betterproto.message_field(1)
    duration: Optional[int] = betterproto.message_field(
        2, wraps=betterproto.TYPE_UINT32
    )


@dataclass(eq=False, repr=False)
class QuaProgramRampPulse(betterproto.Message):
    loc: str = betterproto.string_field(100)
    value: "QuaProgramAnyScalarExpression" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class QuaProgramChirp(betterproto.Message):
    loc: str = betterproto.string_field(100)
    scalar_rate: "QuaProgramAnyScalarExpression" = betterproto.message_field(
        1, group="rate"
    )
    array_rate: "QuaProgramArrayVarRefExpression" = betterproto.message_field(
        2, group="rate"
    )
    units: "QuaProgramChirpUnits" = betterproto.enum_field(3)
    times: List[int] = betterproto.uint32_field(4)
    continue_chirp: bool = betterproto.bool_field(5)


@dataclass(eq=False, repr=False)
class QuaProgramPlayStatement(betterproto.Message):
    loc: str = betterproto.string_field(100)
    qe: "QuaProgramQuantumElementReference" = betterproto.message_field(1)
    pulse: "QuaProgramPulseReference" = betterproto.message_field(2)
    named_pulse: "QuaProgramPulseReference" = betterproto.message_field(
        7, group="pulseType"
    )
    ramp_pulse: "QuaProgramRampPulse" = betterproto.message_field(8, group="pulseType")
    amp: "QuaProgramAmpMultiplier" = betterproto.message_field(3)
    duration: "QuaProgramAnyScalarExpression" = betterproto.message_field(4)
    condition: "QuaProgramAnyScalarExpression" = betterproto.message_field(5)
    port_condition: "QuaProgramAnyScalarExpression" = betterproto.message_field(12)
    target_input: str = betterproto.string_field(6)
    chirp: "QuaProgramChirp" = betterproto.message_field(9)
    truncate: "QuaProgramAnyScalarExpression" = betterproto.message_field(10)
    timestamp_label: str = betterproto.string_field(11)

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.is_set("pulse"):
            warnings.warn(
                "QuaProgramPlayStatement.pulse is deprecated", DeprecationWarning
            )


@dataclass(eq=False, repr=False)
class QuaProgramUpdateFrequencyStatement(betterproto.Message):
    loc: str = betterproto.string_field(100)
    qe: "QuaProgramQuantumElementReference" = betterproto.message_field(1)
    value: "QuaProgramAnyScalarExpression" = betterproto.message_field(2)
    units: "QuaProgramUpdateFrequencyStatementUnits" = betterproto.enum_field(3)
    keep_phase: bool = betterproto.bool_field(4)


@dataclass(eq=False, repr=False)
class QuaProgramUpdateOscillatorFrequencyStatement(betterproto.Message):
    loc: str = betterproto.string_field(100)
    qe: "QuaProgramQuantumElementReference" = betterproto.message_field(1)
    value: "QuaProgramAnyScalarExpression" = betterproto.message_field(2)
    units: "QuaProgramUpdateOscillatorFrequencyStatementUnits" = betterproto.enum_field(
        3
    )
    keep_phase: bool = betterproto.bool_field(4)
    update_component: "QuaProgramUpdateOscillatorFrequencyStatementUpdateComponentSelection" = betterproto.enum_field(
        5
    )


@dataclass(eq=False, repr=False)
class QuaProgramSetDcOffsetStatement(betterproto.Message):
    loc: str = betterproto.string_field(100)
    qe: "QuaProgramQuantumElementReference" = betterproto.message_field(1)
    qe_input_reference: str = betterproto.string_field(2)
    offset: "QuaProgramAnyScalarExpression" = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class QuaProgramZRotationStatement(betterproto.Message):
    loc: str = betterproto.string_field(100)
    qe: "QuaProgramQuantumElementReference" = betterproto.message_field(3)
    value: "QuaProgramAnyScalarExpression" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class QuaProgramResetFrameStatement(betterproto.Message):
    loc: str = betterproto.string_field(100)
    qe: "QuaProgramQuantumElementReference" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class QuaProgramFastFrameRotationStatement(betterproto.Message):
    loc: str = betterproto.string_field(100)
    qe: "QuaProgramQuantumElementReference" = betterproto.message_field(1)
    cosine: "QuaProgramAnyScalarExpression" = betterproto.message_field(2)
    sine: "QuaProgramAnyScalarExpression" = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class QuaProgramAlignStatement(betterproto.Message):
    loc: str = betterproto.string_field(100)
    qe: List["QuaProgramQuantumElementReference"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class QuaProgramMeasureProcess(betterproto.Message):
    analog: "QuaProgramAnalogMeasureProcess" = betterproto.message_field(
        1, group="process"
    )
    digital: "QuaProgramDigitalMeasureProcess" = betterproto.message_field(
        2, group="process"
    )


@dataclass(eq=False, repr=False)
class QuaProgramAnalogMeasureProcess(betterproto.Message):
    loc: str = betterproto.string_field(100)
    bare_integration: "QuaProgramAnalogMeasureProcessBareIntegration" = (
        betterproto.message_field(1, group="process")
    )
    demod_integration: "QuaProgramAnalogMeasureProcessDemodIntegration" = (
        betterproto.message_field(2, group="process")
    )
    raw_time_tagging: "QuaProgramAnalogMeasureProcessRawTimeTagging" = (
        betterproto.message_field(3, group="process")
    )
    dual_bare_integration: "QuaProgramAnalogMeasureProcessDualBareIntegration" = (
        betterproto.message_field(4, group="process")
    )
    dual_demod_integration: "QuaProgramAnalogMeasureProcessDualDemodIntegration" = (
        betterproto.message_field(5, group="process")
    )
    high_res_time_tagging: "QuaProgramAnalogMeasureProcessHighResTimeTagging" = (
        betterproto.message_field(6, group="process")
    )


@dataclass(eq=False, repr=False)
class QuaProgramAnalogMeasureProcessBareIntegration(betterproto.Message):
    integration: "QuaProgramIntegrationWeightReference" = betterproto.message_field(1)
    target: "QuaProgramAnalogProcessTarget" = betterproto.message_field(2)
    element_output: str = betterproto.string_field(9)


@dataclass(eq=False, repr=False)
class QuaProgramAnalogMeasureProcessDualBareIntegration(betterproto.Message):
    integration1: "QuaProgramIntegrationWeightReference" = betterproto.message_field(1)
    integration2: "QuaProgramIntegrationWeightReference" = betterproto.message_field(2)
    target: "QuaProgramAnalogProcessTarget" = betterproto.message_field(3)
    element_output1: str = betterproto.string_field(9)
    element_output2: str = betterproto.string_field(10)


@dataclass(eq=False, repr=False)
class QuaProgramAnalogMeasureProcessDemodIntegration(betterproto.Message):
    integration: "QuaProgramIntegrationWeightReference" = betterproto.message_field(1)
    target: "QuaProgramAnalogProcessTarget" = betterproto.message_field(2)
    element_output: str = betterproto.string_field(9)


@dataclass(eq=False, repr=False)
class QuaProgramAnalogMeasureProcessDualDemodIntegration(betterproto.Message):
    integration1: "QuaProgramIntegrationWeightReference" = betterproto.message_field(1)
    integration2: "QuaProgramIntegrationWeightReference" = betterproto.message_field(2)
    target: "QuaProgramAnalogProcessTarget" = betterproto.message_field(3)
    element_output1: str = betterproto.string_field(9)
    element_output2: str = betterproto.string_field(10)


@dataclass(eq=False, repr=False)
class QuaProgramAnalogMeasureProcessRawTimeTagging(betterproto.Message):
    target: "QuaProgramArrayVarRefExpression" = betterproto.message_field(1)
    target_len: "QuaProgramVarRefExpression" = betterproto.message_field(2)
    max_time: int = betterproto.int32_field(3)
    element_output: str = betterproto.string_field(9)


@dataclass(eq=False, repr=False)
class QuaProgramAnalogMeasureProcessHighResTimeTagging(betterproto.Message):
    target: "QuaProgramArrayVarRefExpression" = betterproto.message_field(1)
    target_len: "QuaProgramVarRefExpression" = betterproto.message_field(2)
    max_time: int = betterproto.int32_field(3)
    element_output: str = betterproto.string_field(9)


@dataclass(eq=False, repr=False)
class QuaProgramDigitalMeasureProcess(betterproto.Message):
    loc: str = betterproto.string_field(100)
    raw_time_tagging: "QuaProgramDigitalMeasureProcessRawTimeTagging" = (
        betterproto.message_field(1, group="process")
    )
    counting: "QuaProgramDigitalMeasureProcessCounting" = betterproto.message_field(
        2, group="process"
    )


@dataclass(eq=False, repr=False)
class QuaProgramDigitalMeasureProcessRawTimeTagging(betterproto.Message):
    element_output: str = betterproto.string_field(1)
    target: "QuaProgramArrayVarRefExpression" = betterproto.message_field(2)
    target_len: "QuaProgramVarRefExpression" = betterproto.message_field(3)
    max_time: int = betterproto.int32_field(4)


@dataclass(eq=False, repr=False)
class QuaProgramDigitalMeasureProcessCounting(betterproto.Message):
    element_outputs: List[str] = betterproto.string_field(1)
    target: "QuaProgramVarRefExpression" = betterproto.message_field(2)
    max_time: int = betterproto.int32_field(3)


@dataclass(eq=False, repr=False)
class QuaProgramAnalogProcessTarget(betterproto.Message):
    loc: str = betterproto.string_field(100)
    scalar_process: "QuaProgramAnalogProcessTargetScalarProcessTarget" = (
        betterproto.message_field(1, group="processTarget")
    )
    vector_process: "QuaProgramAnalogProcessTargetVectorProcessTarget" = (
        betterproto.message_field(2, group="processTarget")
    )


@dataclass(eq=False, repr=False)
class QuaProgramAnalogProcessTargetScalarProcessTarget(betterproto.Message):
    variable: "QuaProgramVarRefExpression" = betterproto.message_field(
        1, group="target"
    )
    array_cell: "QuaProgramArrayCellRefExpression" = betterproto.message_field(
        2, group="target"
    )


@dataclass(eq=False, repr=False)
class QuaProgramAnalogProcessTargetVectorProcessTarget(betterproto.Message):
    array: "QuaProgramArrayVarRefExpression" = betterproto.message_field(1)
    time_division: "QuaProgramAnalogProcessTargetTimeDivision" = (
        betterproto.message_field(2)
    )


@dataclass(eq=False, repr=False)
class QuaProgramAnalogProcessTargetTimeDivision(betterproto.Message):
    sliced: "QuaProgramAnalogTimeDivisionSliced" = betterproto.message_field(
        2, group="timeDivision"
    )
    accumulated: "QuaProgramAnalogTimeDivisionAccumulated" = betterproto.message_field(
        3, group="timeDivision"
    )
    moving_window: "QuaProgramAnalogTimeDivisionMovingWindow" = (
        betterproto.message_field(4, group="timeDivision")
    )


@dataclass(eq=False, repr=False)
class QuaProgramAnalogTimeDivision(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class QuaProgramAnalogTimeDivisionSliced(betterproto.Message):
    loc: str = betterproto.string_field(100)
    samples_per_chunk: int = betterproto.int32_field(1)


@dataclass(eq=False, repr=False)
class QuaProgramAnalogTimeDivisionAccumulated(betterproto.Message):
    loc: str = betterproto.string_field(100)
    samples_per_chunk: int = betterproto.int32_field(1)


@dataclass(eq=False, repr=False)
class QuaProgramAnalogTimeDivisionMovingWindow(betterproto.Message):
    loc: str = betterproto.string_field(100)
    samples_per_chunk: int = betterproto.int32_field(1)
    chunks_per_window: int = betterproto.int32_field(2)


@dataclass(eq=False, repr=False)
class QuaProgramMeasureOutput(betterproto.Message):
    loc: str = betterproto.string_field(100)
    integration: "QuaProgramIntegrationWeightReference" = betterproto.message_field(1)
    variable: "QuaProgramVarRefExpression" = betterproto.message_field(2)
    output: str = betterproto.string_field(3)
    target: "QuaProgramMeasureOutputTarget" = betterproto.message_field(4)

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.is_set("variable"):
            warnings.warn(
                "QuaProgramMeasureOutput.variable is deprecated", DeprecationWarning
            )


@dataclass(eq=False, repr=False)
class QuaProgramMeasureOutputTarget(betterproto.Message):
    variable: "QuaProgramVarRefExpression" = betterproto.message_field(
        1, group="target"
    )
    array_cell: "QuaProgramArrayCellRefExpression" = betterproto.message_field(
        2, group="target"
    )


@dataclass(eq=False, repr=False)
class QuaProgramMeasureStatement(betterproto.Message):
    loc: str = betterproto.string_field(100)
    qe: "QuaProgramQuantumElementReference" = betterproto.message_field(1)
    pulse: "QuaProgramPulseReference" = betterproto.message_field(2)
    amp: "QuaProgramAmpMultiplier" = betterproto.message_field(3)
    outputs: List["QuaProgramMeasureOutput"] = betterproto.message_field(4)
    stream_as: str = betterproto.string_field(5)
    measure_processes: List["QuaProgramMeasureProcess"] = betterproto.message_field(7)
    timestamp_label: str = betterproto.string_field(8)

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.is_set("outputs"):
            warnings.warn(
                "QuaProgramMeasureStatement.outputs is deprecated", DeprecationWarning
            )


@dataclass(eq=False, repr=False)
class QuaProgramWaitStatement(betterproto.Message):
    loc: str = betterproto.string_field(100)
    qe: List["QuaProgramQuantumElementReference"] = betterproto.message_field(1)
    time: "QuaProgramAnyScalarExpression" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class QuaProgramWaitForTriggerStatement(betterproto.Message):
    loc: str = betterproto.string_field(100)
    qe: List["QuaProgramQuantumElementReference"] = betterproto.message_field(1)
    pulse_to_play: "QuaProgramPulseReference" = betterproto.message_field(2)
    global_trigger: "betterproto_lib_google_protobuf.Empty" = betterproto.message_field(
        3, group="source"
    )
    element_output: "QuaProgramWaitForTriggerStatementElementOutput" = (
        betterproto.message_field(4, group="source")
    )
    time_tag_target: "QuaProgramVarRefExpression" = betterproto.message_field(5)


@dataclass(eq=False, repr=False)
class QuaProgramWaitForTriggerStatementElementOutput(betterproto.Message):
    element: str = betterproto.string_field(1)
    output: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class QuaProgramUpdateCorrectionStatement(betterproto.Message):
    loc: str = betterproto.string_field(100)
    qe: "QuaProgramQuantumElementReference" = betterproto.message_field(1)
    correction: "QuaProgramCorrection" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class QuaProgramResetPhaseStatement(betterproto.Message):
    loc: str = betterproto.string_field(100)
    qe: "QuaProgramQuantumElementReference" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class QuaProgramResetGlobalPhaseStatement(betterproto.Message):
    loc: str = betterproto.string_field(100)


@dataclass(eq=False, repr=False)
class QuaProgramCorrection(betterproto.Message):
    loc: str = betterproto.string_field(100)
    c0: "QuaProgramAnyScalarExpression" = betterproto.message_field(1)
    c1: "QuaProgramAnyScalarExpression" = betterproto.message_field(2)
    c2: "QuaProgramAnyScalarExpression" = betterproto.message_field(3)
    c3: "QuaProgramAnyScalarExpression" = betterproto.message_field(4)


@dataclass(eq=False, repr=False)
class QuaProgramSyncStatement(betterproto.Message):
    loc: str = betterproto.string_field(100)


@dataclass(eq=False, repr=False)
class QuaProgramElseIf(betterproto.Message):
    condition: "QuaProgramAnyScalarExpression" = betterproto.message_field(1)
    body: "QuaProgramStatementsCollection" = betterproto.message_field(2)
    loc: str = betterproto.string_field(100)


@dataclass(eq=False, repr=False)
class QuaProgramIfStatement(betterproto.Message):
    condition: "QuaProgramAnyScalarExpression" = betterproto.message_field(1)
    body: "QuaProgramStatementsCollection" = betterproto.message_field(2)
    else_: "QuaProgramStatementsCollection" = betterproto.message_field(3)
    elseifs: List["QuaProgramElseIf"] = betterproto.message_field(4)
    unsafe: bool = betterproto.bool_field(5)
    loc: str = betterproto.string_field(100)


@dataclass(eq=False, repr=False)
class QuaProgramForStatement(betterproto.Message):
    init: "QuaProgramStatementsCollection" = betterproto.message_field(1)
    condition: "QuaProgramAnyScalarExpression" = betterproto.message_field(2)
    update: "QuaProgramStatementsCollection" = betterproto.message_field(3)
    body: "QuaProgramStatementsCollection" = betterproto.message_field(4)
    loc: str = betterproto.string_field(100)


@dataclass(eq=False, repr=False)
class QuaProgramQRunStatement(betterproto.Message):
    body: "QuaProgramStatementsCollection" = betterproto.message_field(1)
    loc: str = betterproto.string_field(100)


@dataclass(eq=False, repr=False)
class QuaProgramStrictTimingStatement(betterproto.Message):
    body: "QuaProgramStatementsCollection" = betterproto.message_field(1)
    loc: str = betterproto.string_field(100)


@dataclass(eq=False, repr=False)
class QuaProgramForEachStatement(betterproto.Message):
    iterator: List[
        "QuaProgramForEachStatementVariableWithValues"
    ] = betterproto.message_field(1)
    body: "QuaProgramStatementsCollection" = betterproto.message_field(2)
    loc: str = betterproto.string_field(100)


@dataclass(eq=False, repr=False)
class QuaProgramForEachStatementVariableWithValues(betterproto.Message):
    variable: "QuaProgramVarRefExpression" = betterproto.message_field(1)
    array: "QuaProgramArrayVarRefExpression" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class QuaProgramRusStatement(betterproto.Message):
    do_block: "QuaProgramStatementsCollection" = betterproto.message_field(1)
    fail_block: "QuaProgramStatementsCollection" = betterproto.message_field(2)
    until: "QuaProgramAnyScalarExpression" = betterproto.message_field(3)
    loc: str = betterproto.string_field(100)


@dataclass(eq=False, repr=False)
class QuaProgramAssignmentStatement(betterproto.Message):
    variable: "QuaProgramVarRefExpression" = betterproto.message_field(1)
    expression: "QuaProgramAnyScalarExpression" = betterproto.message_field(2)
    save: bool = betterproto.bool_field(3)
    target: "QuaProgramAssignmentStatementTarget" = betterproto.message_field(4)
    loc: str = betterproto.string_field(100)

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.is_set("variable"):
            warnings.warn(
                "QuaProgramAssignmentStatement.variable is deprecated",
                DeprecationWarning,
            )
        if self.is_set("save"):
            warnings.warn(
                "QuaProgramAssignmentStatement.save is deprecated", DeprecationWarning
            )


@dataclass(eq=False, repr=False)
class QuaProgramAssignmentStatementTarget(betterproto.Message):
    variable: "QuaProgramVarRefExpression" = betterproto.message_field(
        1, group="target"
    )
    array_cell: "QuaProgramArrayCellRefExpression" = betterproto.message_field(
        2, group="target"
    )


@dataclass(eq=False, repr=False)
class QuaProgramPauseStatement(betterproto.Message):
    qes: List["QuaProgramQuantumElementReference"] = betterproto.message_field(1)
    loc: str = betterproto.string_field(100)


@dataclass(eq=False, repr=False)
class QuaProgramSaveStatement(betterproto.Message):
    variable: "QuaProgramVarRefExpression" = betterproto.message_field(1)
    tag: str = betterproto.string_field(2)
    source: "QuaProgramSaveStatementSource" = betterproto.message_field(3)
    loc: str = betterproto.string_field(100)

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.is_set("variable"):
            warnings.warn(
                "QuaProgramSaveStatement.variable is deprecated", DeprecationWarning
            )


@dataclass(eq=False, repr=False)
class QuaProgramSaveStatementSource(betterproto.Message):
    variable: "QuaProgramVarRefExpression" = betterproto.message_field(
        1, group="source"
    )
    array_cell: "QuaProgramArrayCellRefExpression" = betterproto.message_field(
        2, group="source"
    )
    literal: "QuaProgramLiteralExpression" = betterproto.message_field(
        3, group="source"
    )


@dataclass(eq=False, repr=False)
class QuaProgramStatementsCollection(betterproto.Message):
    statements: List["QuaProgramAnyStatement"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class QuaProgramAnyScalarExpression(betterproto.Message):
    variable: "QuaProgramVarRefExpression" = betterproto.message_field(
        1, group="expression_oneof"
    )
    literal: "QuaProgramLiteralExpression" = betterproto.message_field(
        2, group="expression_oneof"
    )
    binary_operation: "QuaProgramBinaryExpression" = betterproto.message_field(
        3, group="expression_oneof"
    )
    array_cell: "QuaProgramArrayCellRefExpression" = betterproto.message_field(
        4, group="expression_oneof"
    )
    array_length: "QuaProgramArrayLengthExpression" = betterproto.message_field(
        5, group="expression_oneof"
    )
    lib_function: "QuaProgramLibFunctionExpression" = betterproto.message_field(
        6, group="expression_oneof"
    )


@dataclass(eq=False, repr=False)
class QuaProgramLibFunctionExpression(betterproto.Message):
    function_name: str = betterproto.string_field(1)
    arguments: List[
        "QuaProgramLibFunctionExpressionArgument"
    ] = betterproto.message_field(2)
    library_name: str = betterproto.string_field(3)
    loc: str = betterproto.string_field(100)


@dataclass(eq=False, repr=False)
class QuaProgramLibFunctionExpressionArgument(betterproto.Message):
    scalar: "QuaProgramAnyScalarExpression" = betterproto.message_field(
        1, group="argument_oneof"
    )
    array: "QuaProgramArrayVarRefExpression" = betterproto.message_field(
        2, group="argument_oneof"
    )


@dataclass(eq=False, repr=False)
class QuaProgramAdvanceInputStreamStatement(betterproto.Message):
    stream_variable: "QuaProgramVarRefExpression" = betterproto.message_field(
        1, group="stream_oneof"
    )
    stream_array: "QuaProgramArrayVarRefExpression" = betterproto.message_field(
        2, group="stream_oneof"
    )
    loc: str = betterproto.string_field(100)


@dataclass(eq=False, repr=False)
class QuaProgramArrayLengthExpression(betterproto.Message):
    array: "QuaProgramArrayVarRefExpression" = betterproto.message_field(1)
    loc: str = betterproto.string_field(100)


@dataclass(eq=False, repr=False)
class QuaProgramVarRefExpression(betterproto.Message):
    name: str = betterproto.string_field(2, group="var_oneof")
    io_number: int = betterproto.int32_field(3, group="var_oneof")
    loc: str = betterproto.string_field(100)


@dataclass(eq=False, repr=False)
class QuaProgramArrayVarRefExpression(betterproto.Message):
    name: str = betterproto.string_field(2)
    loc: str = betterproto.string_field(100)


@dataclass(eq=False, repr=False)
class QuaProgramArrayCellRefExpression(betterproto.Message):
    array_var: "QuaProgramArrayVarRefExpression" = betterproto.message_field(2)
    index: "QuaProgramAnyScalarExpression" = betterproto.message_field(3)
    loc: str = betterproto.string_field(100)


@dataclass(eq=False, repr=False)
class QuaProgramLiteralExpression(betterproto.Message):
    value: str = betterproto.string_field(1)
    type: "QuaProgramType" = betterproto.enum_field(2)
    loc: str = betterproto.string_field(100)


@dataclass(eq=False, repr=False)
class QuaProgramBinaryExpression(betterproto.Message):
    op: "QuaProgramBinaryExpressionBinaryOperator" = betterproto.enum_field(1)
    left: "QuaProgramAnyScalarExpression" = betterproto.message_field(4)
    right: "QuaProgramAnyScalarExpression" = betterproto.message_field(5)
    loc: str = betterproto.string_field(100)


@dataclass(eq=False, repr=False)
class QuaProgramDynamicConfig(betterproto.Message):
    """
    A config that is similar to JSON in structure and can be converted to the real config
    It has support for versioning
    """

    version: int = betterproto.uint32_field(1)
    root: "betterproto_lib_google_protobuf.Struct" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class QuaResultAnalysis(betterproto.Message):
    version: int = betterproto.uint32_field(1)
    model: List[
        "betterproto_lib_google_protobuf.ListValue"
    ] = betterproto.message_field(3)



import warnings
from typing import List, Optional
from dataclasses import field, dataclass

from qm.grpc.qua import QuaProgramCompilerOptions


@dataclass
class CompilerOptionArguments:
    strict: Optional[bool] = field(default=None)

    flags: List[str] = field(default_factory=list)


def get_request_compiler_options(
    compiler_options: CompilerOptionArguments,
) -> QuaProgramCompilerOptions:
    flags = compiler_options.flags
    if compiler_options.strict:
        flags.append("strict")

    return QuaProgramCompilerOptions(flags=compiler_options.flags)


def standardize_compiler_params(
    compiler_options: Optional[CompilerOptionArguments],
    strict: Optional[bool],
    flags: Optional[List[str]],
) -> CompilerOptionArguments:
    if compiler_options is not None:
        if (strict is not None) or (flags is not None):
            raise ValueError("Please remove **kwargs ('flags' and 'strict') from calling to 'simulate'")
        return compiler_options

    if (strict is not None) or (flags is not None):
        warnings.warn(
            "Using **kwargs for the compiler arguments is deprecated, and will be removed in 1.3.0. "
            "Please set the options inside the object 'CompilerOptionArguments'.",
            DeprecationWarning,
            stacklevel=2,
        )
        return CompilerOptionArguments(strict=strict, flags=flags or [])

    return CompilerOptionArguments()



import logging
from typing import Union, Optional, cast

import numpy.typing

from qm.exceptions import QmInvalidSchemaError
from qm.results.base_streaming_result_fetcher import BaseStreamingResultFetcher

logger = logging.getLogger(__name__)


class SingleStreamingResultFetcher(BaseStreamingResultFetcher):
    """A handle to a result of a pipeline terminating with ``save``"""

    def _validate_schema(self) -> None:
        if not self._schema.is_single:
            raise QmInvalidSchemaError("expecting a single-result schema")

    def wait_for_values(self, count: int = 1, timeout: float = float("inf")) -> None:
        if count != 1:
            raise RuntimeError("single result can wait only for a single value")
        super().wait_for_values(1, timeout)

    def fetch_all(
        self, *, check_for_errors: bool = True, flat_struct: bool = False
    ) -> Optional[numpy.typing.NDArray[numpy.generic]]:
        """Fetch a result from the current result stream saved in server memory.
        The result stream is populated by the save() and save_all() statements.
        Note that if save_all() statements are used, calling this function twice
        may give different results.

        Args:
            check_for_errors: If true, the function would also check whether run-time errors happened during the
                program execution and would write to the logger an error message.
            flat_struct: results will have a flat structure - dimensions will be part of the shape and not of the type

        Returns:
            all result of current result stream
        """
        return self.fetch(0, flat_struct=flat_struct, check_for_errors=check_for_errors)

    def fetch(
        self,
        item: Union[int, slice],
        *,
        check_for_errors: bool = True,
        flat_struct: bool = False,
    ) -> Optional[numpy.typing.NDArray[numpy.generic]]:
        """Fetch a single result from the current result stream saved in server memory.
        The result stream is populated by the save().

        Args:
            item: ignored
            check_for_errors: If true, the function would also check whether run-time errors happened during the
                program execution and would write to the logger an error message.
            flat_struct: results will have a flat structure - dimensions will be part of the shape and not of the type

        Returns:
            the current result

        Example:
            ```python
            res.fetch() # return the item in the top position
            ```
        """
        if (isinstance(item, int) and item != 0) or isinstance(item, slice):
            logger.warning("Fetching single result will always return the single value")
        value = self.strict_fetch(0, check_for_errors=check_for_errors, flat_struct=flat_struct)
        if flat_struct:
            if len(value) == 0:
                logger.warning("Nothing to fetch: no results were found. Please wait until the results are ready.")
                return None
            elif len(value) == 1:
                return cast(numpy.typing.NDArray[numpy.generic], value[0])
            else:
                return value
        else:
            if len(value) == 0:
                logger.warning("Nothing to fetch: no results were found. Please wait until the results are ready.")
                return None
            elif len(value[0]) == 1:
                return cast(numpy.typing.NDArray[numpy.generic], value[0][0])
            else:
                return cast(numpy.typing.NDArray[numpy.generic], value[0])



from qm.results.streaming_result_fetcher import StreamingResultFetcher
from qm.results.single_streaming_result_fetcher import SingleStreamingResultFetcher
from qm.results.multiple_streaming_result_fetcher import MultipleStreamingResultFetcher

__all__ = ["StreamingResultFetcher", "SingleStreamingResultFetcher", "MultipleStreamingResultFetcher"]



from typing import Set, Dict, List, Optional, TypedDict


class PulserLocationType(TypedDict, total=False):
    controllerName: str
    pulserIndex: int
    femId: int


class IqInfoType(TypedDict, total=False):
    isPartOfIq: bool
    iqGroupId: int
    isI: bool
    isQ: bool


class ChirpInfoType(TypedDict, total=False):
    rate: List[int]
    times: List[int]
    units: str
    startFrequency: float
    endFrequency: float


class PlayedWaveformType(TypedDict, total=False):
    waveformName: str
    pulseName: str
    pulser: PulserLocationType
    iqInfo: IqInfoType
    timestamp: int
    length: int
    endsAt: int
    outputPorts: Set[int]
    quantumElements: str


class PortToDucMap(TypedDict):
    port: float
    ducs: List[float]


class PlayedAnalogWaveformType(PlayedWaveformType, total=False):
    currentFrame: List[float]
    currentCorrectionElements: List[float]
    currentIntermediateFrequency: float
    currentGMatrixElements: List[float]
    currentDCOffsetByPort: Dict[int, float]
    currentPhase: float
    chirpInfo: ChirpInfoType
    portToDuc: List[PortToDucMap]


class AdcAcquisitionType(TypedDict, total=False):
    startTime: int
    endTime: int
    process: str
    pulser: PulserLocationType
    quantumElement: str
    adc: List[int]


class WaveformReportType(TypedDict, total=False):
    analogWaveforms: List[PlayedAnalogWaveformType]
    digitalWaveforms: List[PlayedWaveformType]
    adcAcquisitions: List[AdcAcquisitionType]


class WaveformPlayingType(TypedDict, total=False):
    name: str
    timestamp: int
    duration: int
    frequency: float
    phase: float


class WaveformInControllerType(TypedDict):
    ports: Dict[int, List[WaveformPlayingType]]


class WaveformInPortsType(TypedDict):
    controllers: Dict[str, WaveformInControllerType]
    elements: Dict[str, WaveformPlayingType]


class AnalogOutputsType(TypedDict):
    waveforms: Optional[WaveformInPortsType]


class DigitalOutputsType(TypedDict):
    waveforms: Optional[WaveformInPortsType]



import traceback
from pathlib import Path

from qm.exceptions import QmQuaException

QM_PACKAGE_DIR = Path(__file__).resolve().parent


def _get_loc() -> str:
    tb = traceback.extract_stack()
    for trace in reversed(tb):
        if QM_PACKAGE_DIR not in Path(trace.filename).parents:
            return f'File "{trace.filename}", line {trace.lineno}: {trace.line} '
    raise QmQuaException("Could not find location")



from abc import ABCMeta, abstractmethod
from typing import Tuple, Union, Mapping, Optional

from qm.octave.calibration_utils import Correction
from qm.octave.octave_mixer_calibration import MixerCalibrationResults


class AbstractLOCalibration(metaclass=ABCMeta):
    @abstractmethod
    def get_i0(self) -> float:
        pass

    @abstractmethod
    def get_q0(self) -> float:
        pass


class AbstractIFCalibration(metaclass=ABCMeta):
    @abstractmethod
    def get_correction(self) -> Correction:
        pass


class AbstractCalibrationDB(metaclass=ABCMeta):
    """
    Abstract class for the calibration database.
    """

    @abstractmethod
    def get_lo_cal(
        self, octave_channel: Tuple[str, int], lo_freq: Union[int, float], gain: Optional[float]
    ) -> Optional[AbstractLOCalibration]:
        """
        Get the LO calibration for a given octave channel, LO frequency, and gain.
        Args:
            octave_channel: The octave channel to get the calibration for.
            lo_freq: The LO frequency to get the calibration for.
            gain: The gain to get the calibration for.
        Returns:
            The LO calibration for the given parameters, or None if it doesn't exist.
        """
        pass

    @abstractmethod
    def get_if_cal(
        self,
        octave_channel: Tuple[str, int],
        lo_freq: Union[int, float],
        gain: Optional[float],
        if_freq: Union[int, float],
    ) -> Optional[AbstractIFCalibration]:
        """
        Get the IF calibration for a given octave channel, LO frequency, gain, and IF frequency.
        Args:
            octave_channel: The octave channel to get the calibration for.
            lo_freq: The LO frequency to get the calibration for.
            gain: The gain to get the calibration for.
            if_freq: The IF frequency to get the calibration for.
        Returns:
            The IF calibration for the given parameters, or None if it doesn't exist.
        """
        pass

    @abstractmethod
    def get_all_if_cal_for_lo(
        self, octave_channel: Tuple[str, int], lo_freq: Union[int, float], gain: Optional[float]
    ) -> Mapping[Union[int, float], AbstractIFCalibration]:
        """
        Get IF calibration for all IF frequencies for a given octave channel, LO frequency, and gain.
        Args:
            octave_channel: The octave channel to get the calibration for.
            lo_freq: The LO frequency to get the calibration for.
            gain: The gain to get the calibration for.
        Returns:
            A dictionary of IF frequencies to their respective IF calibration, or an empty dictionary if none exist.
        """
        pass

    @abstractmethod
    def update_calibration_result(
        self, result: MixerCalibrationResults, octave_channel: Tuple[str, int], method: str = ""
    ) -> None:
        """
        Update the calibration database with the given calibration results.
        Args:
            result: The calibration results to update the database with.
            octave_channel: The octave channel to update the database with.
            method: Deprecated.
        """
        pass



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: qm/pb/job_results.proto
# plugin: python-betterproto
# This file has been @generated
import warnings
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    AsyncIterator,
    Dict,
    List,
    Optional,
)

import betterproto
import betterproto.lib.google.protobuf as betterproto_lib_google_protobuf
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase

from .. import general_messages as _general_messages__


if TYPE_CHECKING:
    import grpclib.server
    from betterproto.grpc.grpclib_client import MetadataLike
    from grpclib.metadata import Deadline


class GetJobErrorsResponseExecutionErrorSeverity(betterproto.Enum):
    ERROR = 0
    WARNING = 1


@dataclass(eq=False, repr=False)
class GetProgramMetadataRequest(betterproto.Message):
    job_id: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class GetProgramMetadataResponse(betterproto.Message):
    success: bool = betterproto.bool_field(1)
    job_id: str = betterproto.string_field(2)
    program_stream_metadata: "ProgramStreamMetadata" = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class ProgramStreamMetadata(betterproto.Message):
    stream_metadata: List["StreamMetadata"] = betterproto.message_field(1)
    stream_metadata_extraction_error: List[
        "StreamMetadataExtractionError"
    ] = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class StreamMetadata(betterproto.Message):
    stream_name: str = betterproto.string_field(1)
    iteration_data: List["IterationData"] = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class StreamMetadataExtractionError(betterproto.Message):
    location: str = betterproto.string_field(1)
    error: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class IterationData(betterproto.Message):
    iteration_variable_name: str = betterproto.string_field(1)
    for_each_int_iteration_values: "IterationDataForEachIntIterationValues" = (
        betterproto.message_field(2, group="iterationValues")
    )
    for_int_iteration_values: "IterationDataForIntIterationValues" = (
        betterproto.message_field(3, group="iterationValues")
    )
    for_each_double_iteration_values: "IterationDataForEachDoubleIterationValues" = (
        betterproto.message_field(4, group="iterationValues")
    )
    for_double_iteration_values: "IterationDataForDoubleIterationValues" = (
        betterproto.message_field(5, group="iterationValues")
    )


@dataclass(eq=False, repr=False)
class IterationDataForEachIntIterationValues(betterproto.Message):
    values: List[int] = betterproto.int32_field(1)


@dataclass(eq=False, repr=False)
class IterationDataForIntIterationValues(betterproto.Message):
    start_value: int = betterproto.int32_field(1)
    step: int = betterproto.int32_field(2)
    number_of_iterations: int = betterproto.int32_field(3)


@dataclass(eq=False, repr=False)
class IterationDataForEachDoubleIterationValues(betterproto.Message):
    values: List[float] = betterproto.double_field(1)


@dataclass(eq=False, repr=False)
class IterationDataForDoubleIterationValues(betterproto.Message):
    start_value: float = betterproto.double_field(1)
    step: float = betterproto.double_field(2)
    number_of_iterations: int = betterproto.int32_field(3)


@dataclass(eq=False, repr=False)
class GetJobResultSchemaRequest(betterproto.Message):
    job_id: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class GetJobResultSchemaResponse(betterproto.Message):
    items: List["GetJobResultSchemaResponseItem"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class GetJobResultSchemaResponseItem(betterproto.Message):
    name: str = betterproto.string_field(1)
    simple_d_type: str = betterproto.string_field(2)
    is_single: bool = betterproto.bool_field(3)
    expected_count: int = betterproto.int32_field(4)
    shape: List[int] = betterproto.int32_field(5)


@dataclass(eq=False, repr=False)
class GetJobStateRequest(betterproto.Message):
    job_id: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class GetJobStateResponse(betterproto.Message):
    done: bool = betterproto.bool_field(1)
    closed: bool = betterproto.bool_field(2)
    has_dataloss: bool = betterproto.bool_field(3)


@dataclass(eq=False, repr=False)
class GetJobNamedResultHeaderRequest(betterproto.Message):
    job_id: str = betterproto.string_field(1)
    output_name: str = betterproto.string_field(2)
    flat_format: bool = betterproto.bool_field(3)


@dataclass(eq=False, repr=False)
class GetJobNamedResultHeaderResponse(betterproto.Message):
    is_single: bool = betterproto.bool_field(1)
    count_so_far: int = betterproto.int32_field(2)
    simple_d_type: str = betterproto.string_field(3)
    done: bool = betterproto.bool_field(4)
    closed: bool = betterproto.bool_field(5)
    has_dataloss: bool = betterproto.bool_field(6)
    shape: List[int] = betterproto.int32_field(7)
    has_execution_errors: Optional[bool] = betterproto.message_field(
        8, wraps=betterproto.TYPE_BOOL
    )

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.is_set("done"):
            warnings.warn(
                "GetJobNamedResultHeaderResponse.done is deprecated", DeprecationWarning
            )
        if self.is_set("closed"):
            warnings.warn(
                "GetJobNamedResultHeaderResponse.closed is deprecated",
                DeprecationWarning,
            )


@dataclass(eq=False, repr=False)
class GetJobNamedResultRequest(betterproto.Message):
    job_id: str = betterproto.string_field(1)
    output_name: str = betterproto.string_field(2)
    offset: int = betterproto.int32_field(3)
    limit: int = betterproto.int32_field(4)
    long_offset: Optional[int] = betterproto.message_field(
        5, wraps=betterproto.TYPE_INT64
    )

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.is_set("offset"):
            warnings.warn(
                "GetJobNamedResultRequest.offset is deprecated", DeprecationWarning
            )


@dataclass(eq=False, repr=False)
class GetJobNamedResultResponse(betterproto.Message):
    count_of_items: int = betterproto.int32_field(1)
    data: bytes = betterproto.bytes_field(2)


@dataclass(eq=False, repr=False)
class GetJobErrorsRequest(betterproto.Message):
    job_id: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class GetJobErrorsResponse(betterproto.Message):
    errors: List["GetJobErrorsResponseError"] = betterproto.message_field(1)
    job_id: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class GetJobErrorsResponseError(betterproto.Message):
    error_code: int = betterproto.int32_field(1)
    error_severity: "GetJobErrorsResponseExecutionErrorSeverity" = (
        betterproto.enum_field(2)
    )
    message: str = betterproto.string_field(3)


@dataclass(eq=False, repr=False)
class PullAnalysedResultsRequest(betterproto.Message):
    job_file_path: str = betterproto.string_field(1)
    metadata: str = betterproto.string_field(2)
    contains_version: bool = betterproto.bool_field(3)


@dataclass(eq=False, repr=False)
class AnalysedResultsResponse(betterproto.Message):
    version: str = betterproto.string_field(1)
    icp_results: List["IcpResultData"] = betterproto.message_field(11)
    stream_results: List["StreamResultData"] = betterproto.message_field(12)
    errors: List["_general_messages__.ErrorMessage"] = betterproto.message_field(13)


@dataclass(eq=False, repr=False)
class PullResultOutputRequest(betterproto.Message):
    job_id: str = betterproto.string_field(1)
    output_name: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class PullResultOutputResponse(betterproto.Message):
    data: "PullResultOutputResponseData" = betterproto.message_field(
        1, group="response"
    )
    header: "PullResultOutputResponseHeader" = betterproto.message_field(
        2, group="response"
    )
    npz: "PullResultOutputResponseNpz" = betterproto.message_field(3, group="response")
    npy: "PullResultOutputResponseNpy" = betterproto.message_field(4, group="response")


@dataclass(eq=False, repr=False)
class PullResultOutputResponseHeader(betterproto.Message):
    version: str = betterproto.string_field(1)
    output_name: str = betterproto.string_field(2)
    data_loss: bool = betterproto.bool_field(3)
    errors: List["_general_messages__.ErrorMessage"] = betterproto.message_field(4)
    single: bool = betterproto.bool_field(5)


@dataclass(eq=False, repr=False)
class PullResultOutputResponseData(betterproto.Message):
    results: "betterproto_lib_google_protobuf.Value" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class PullResultOutputResponseNpz(betterproto.Message):
    npz: bytes = betterproto.bytes_field(1)


@dataclass(eq=False, repr=False)
class PullResultOutputResponseNpy(betterproto.Message):
    name: str = betterproto.string_field(1)
    npy: bytes = betterproto.bytes_field(2)
    count: int = betterproto.int64_field(3)


@dataclass(eq=False, repr=False)
class BinaryWrapper(betterproto.Message):
    data: List[bytes] = betterproto.bytes_field(1)


@dataclass(eq=False, repr=False)
class PullFileResultRequest(betterproto.Message):
    job_id: str = betterproto.string_field(1)
    metadata: str = betterproto.string_field(2)
    as_npz: bool = betterproto.bool_field(4, group="fileType")


@dataclass(eq=False, repr=False)
class FileResultResponse(betterproto.Message):
    job_id: str = betterproto.string_field(1)
    controller_name: str = betterproto.string_field(2)
    group: str = betterproto.string_field(3)
    npz: bytes = betterproto.bytes_field(4, group="file")
    version: str = betterproto.string_field(10)
    errors: List["_general_messages__.ErrorMessage"] = betterproto.message_field(11)


@dataclass(eq=False, repr=False)
class PullSimulatorSamplesRequest(betterproto.Message):
    job_id: str = betterproto.string_field(1)
    include_analog: bool = betterproto.bool_field(2)
    include_digital: bool = betterproto.bool_field(3)
    as_npz: bool = betterproto.bool_field(4, group="fileType")
    include_all_connections: bool = betterproto.bool_field(5)


@dataclass(eq=False, repr=False)
class SimulatorSamplesResponse(betterproto.Message):
    job_id: str = betterproto.string_field(1)
    ok: bool = betterproto.bool_field(5)
    header: "SimulatorSamplesResponseHeader" = betterproto.message_field(
        6, group="body"
    )
    data: "SimulatorSamplesResponseData" = betterproto.message_field(7, group="body")


@dataclass(eq=False, repr=False)
class SimulatorSamplesResponseHeader(betterproto.Message):
    simple_d_type: str = betterproto.string_field(1)
    count_of_items: int = betterproto.int32_field(2)


@dataclass(eq=False, repr=False)
class SimulatorSamplesResponseData(betterproto.Message):
    data: bytes = betterproto.bytes_field(8)


@dataclass(eq=False, repr=False)
class IcpResultData(betterproto.Message):
    name: str = betterproto.string_field(1)
    timestamp: int = betterproto.uint64_field(2)
    data_loss: bool = betterproto.bool_field(3)
    int_value: int = betterproto.int32_field(11, group="data")
    double_value: float = betterproto.double_field(12, group="data")
    boolean_value: bool = betterproto.bool_field(13, group="data")


@dataclass(eq=False, repr=False)
class StreamResultData(betterproto.Message):
    name: str = betterproto.string_field(1)
    timestamp: int = betterproto.uint64_field(2)
    data_loss: bool = betterproto.bool_field(3)
    multiple_sources: bool = betterproto.bool_field(4)
    data_source_name: str = betterproto.string_field(5)
    data: int = betterproto.int64_field(11)


class JobResultsServiceStub(betterproto.ServiceStub):
    async def get_job_result_schema(
        self,
        get_job_result_schema_request: "GetJobResultSchemaRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "GetJobResultSchemaResponse":
        return await self._unary_unary(
            "/qm.grpc.results_analyser.JobResultsService/GetJobResultSchema",
            get_job_result_schema_request,
            GetJobResultSchemaResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def get_job_state(
        self,
        get_job_state_request: "GetJobStateRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "GetJobStateResponse":
        return await self._unary_unary(
            "/qm.grpc.results_analyser.JobResultsService/GetJobState",
            get_job_state_request,
            GetJobStateResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def get_job_named_result_header(
        self,
        get_job_named_result_header_request: "GetJobNamedResultHeaderRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "GetJobNamedResultHeaderResponse":
        return await self._unary_unary(
            "/qm.grpc.results_analyser.JobResultsService/GetJobNamedResultHeader",
            get_job_named_result_header_request,
            GetJobNamedResultHeaderResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def get_job_named_result(
        self,
        get_job_named_result_request: "GetJobNamedResultRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> AsyncIterator[GetJobNamedResultResponse]:
        async for response in self._unary_stream(
            "/qm.grpc.results_analyser.JobResultsService/GetJobNamedResult",
            get_job_named_result_request,
            GetJobNamedResultResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        ):
            yield response

    async def get_job_errors(
        self,
        get_job_errors_request: "GetJobErrorsRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "GetJobErrorsResponse":
        return await self._unary_unary(
            "/qm.grpc.results_analyser.JobResultsService/GetJobErrors",
            get_job_errors_request,
            GetJobErrorsResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def get_program_metadata(
        self,
        get_program_metadata_request: "GetProgramMetadataRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "GetProgramMetadataResponse":
        return await self._unary_unary(
            "/qm.grpc.results_analyser.JobResultsService/GetProgramMetadata",
            get_program_metadata_request,
            GetProgramMetadataResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )


class JobResultsServiceBase(ServiceBase):
    async def get_job_result_schema(
        self, get_job_result_schema_request: "GetJobResultSchemaRequest"
    ) -> "GetJobResultSchemaResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def get_job_state(
        self, get_job_state_request: "GetJobStateRequest"
    ) -> "GetJobStateResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def get_job_named_result_header(
        self, get_job_named_result_header_request: "GetJobNamedResultHeaderRequest"
    ) -> "GetJobNamedResultHeaderResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def get_job_named_result(
        self, get_job_named_result_request: "GetJobNamedResultRequest"
    ) -> AsyncIterator[GetJobNamedResultResponse]:
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)
        yield GetJobNamedResultResponse()

    async def get_job_errors(
        self, get_job_errors_request: "GetJobErrorsRequest"
    ) -> "GetJobErrorsResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def get_program_metadata(
        self, get_program_metadata_request: "GetProgramMetadataRequest"
    ) -> "GetProgramMetadataResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_get_job_result_schema(
        self,
        stream: "grpclib.server.Stream[GetJobResultSchemaRequest, GetJobResultSchemaResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_job_result_schema(request)
        await stream.send_message(response)

    async def __rpc_get_job_state(
        self, stream: "grpclib.server.Stream[GetJobStateRequest, GetJobStateResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_job_state(request)
        await stream.send_message(response)

    async def __rpc_get_job_named_result_header(
        self,
        stream: "grpclib.server.Stream[GetJobNamedResultHeaderRequest, GetJobNamedResultHeaderResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_job_named_result_header(request)
        await stream.send_message(response)

    async def __rpc_get_job_named_result(
        self,
        stream: "grpclib.server.Stream[GetJobNamedResultRequest, GetJobNamedResultResponse]",
    ) -> None:
        request = await stream.recv_message()
        await self._call_rpc_handler_server_stream(
            self.get_job_named_result,
            stream,
            request,
        )

    async def __rpc_get_job_errors(
        self, stream: "grpclib.server.Stream[GetJobErrorsRequest, GetJobErrorsResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_job_errors(request)
        await stream.send_message(response)

    async def __rpc_get_program_metadata(
        self,
        stream: "grpclib.server.Stream[GetProgramMetadataRequest, GetProgramMetadataResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_program_metadata(request)
        await stream.send_message(response)

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/qm.grpc.results_analyser.JobResultsService/GetJobResultSchema": grpclib.const.Handler(
                self.__rpc_get_job_result_schema,
                grpclib.const.Cardinality.UNARY_UNARY,
                GetJobResultSchemaRequest,
                GetJobResultSchemaResponse,
            ),
            "/qm.grpc.results_analyser.JobResultsService/GetJobState": grpclib.const.Handler(
                self.__rpc_get_job_state,
                grpclib.const.Cardinality.UNARY_UNARY,
                GetJobStateRequest,
                GetJobStateResponse,
            ),
            "/qm.grpc.results_analyser.JobResultsService/GetJobNamedResultHeader": grpclib.const.Handler(
                self.__rpc_get_job_named_result_header,
                grpclib.const.Cardinality.UNARY_UNARY,
                GetJobNamedResultHeaderRequest,
                GetJobNamedResultHeaderResponse,
            ),
            "/qm.grpc.results_analyser.JobResultsService/GetJobNamedResult": grpclib.const.Handler(
                self.__rpc_get_job_named_result,
                grpclib.const.Cardinality.UNARY_STREAM,
                GetJobNamedResultRequest,
                GetJobNamedResultResponse,
            ),
            "/qm.grpc.results_analyser.JobResultsService/GetJobErrors": grpclib.const.Handler(
                self.__rpc_get_job_errors,
                grpclib.const.Cardinality.UNARY_UNARY,
                GetJobErrorsRequest,
                GetJobErrorsResponse,
            ),
            "/qm.grpc.results_analyser.JobResultsService/GetProgramMetadata": grpclib.const.Handler(
                self.__rpc_get_program_metadata,
                grpclib.const.Cardinality.UNARY_UNARY,
                GetProgramMetadataRequest,
                GetProgramMetadataResponse,
            ),
        }



from typing import TYPE_CHECKING, List, Union, Optional, cast

import numpy

from qm.persistence import BaseStore
from qm.api.v2.job_result_api import JobResultApi
from qm.api.job_result_api import JobResultServiceApi
from qm.api.models.capabilities import ServerCapabilities
from qm.exceptions import QmNoResultsError, QmInvalidSchemaError
from qm.StreamMetadata import StreamMetadata, StreamMetadataError
from qm.results.base_streaming_result_fetcher import JobResultItemSchema, BaseStreamingResultFetcher

TIMESTAMPS_LEGACY_EXT = "_timestamps"

if TYPE_CHECKING:
    from qm import StreamingResultFetcher


class MultipleStreamingResultFetcher(BaseStreamingResultFetcher):
    """A handle to a result of a pipeline terminating with ``save_all``"""

    def __init__(
        self,
        job_results: "StreamingResultFetcher",
        schema: JobResultItemSchema,
        service: Union[JobResultServiceApi, JobResultApi],
        store: BaseStore,
        stream_metadata_errors: List[StreamMetadataError],
        stream_metadata: Optional[StreamMetadata],
        capabilities: ServerCapabilities,
    ) -> None:
        self.job_results = job_results
        super().__init__(
            schema=schema,
            service=service,
            store=store,
            stream_metadata_errors=stream_metadata_errors,
            stream_metadata=stream_metadata,
            capabilities=capabilities,
        )

    def _validate_schema(self) -> None:
        if self._schema.is_single:
            raise QmInvalidSchemaError("expecting a multi-result schema")

    def fetch(
        self,
        item: Union[int, slice],
        *,
        check_for_errors: bool = True,
        flat_struct: bool = False,
    ) -> Optional[numpy.typing.NDArray[numpy.generic]]:
        """Fetch a result from the current result stream saved in server memory.
        The result stream is populated by the save() and save_all() statements.
        Note that if save_all() statements are used, calling this function twice
        with the same item index may give different results.

        Args:
            item: The index of the result in the saved results stream.
            check_for_errors: If true, the function would also check whether run-time errors happened during the
                program execution and would write to the logger an error message.
            flat_struct: results will have a flat structure - dimensions will be part of the shape and not of the type

        Returns:
            a single result if item is integer or multiple results if item is Python slice object.

        Example:
            ```python
            res.fetch(0)         # return the item in the top position
            res.fetch(1)         # return the item in position number 2
            res.fetch(slice(1,6))# return items from position 1 to position 6 (exclusive)
                                 # same as res.fetch_all()[1:6]
            ```
        """
        if flat_struct:
            return self.strict_fetch(item, check_for_errors=check_for_errors, flat_struct=flat_struct)
        else:
            # legacy support - reconstruct the old structure
            name = self._schema.name
            timestamps_name = name + TIMESTAMPS_LEGACY_EXT
            timestamps_result_handle = self.job_results.get(timestamps_name)
            if timestamps_result_handle is None:
                return self.strict_fetch(item, check_for_errors=check_for_errors)
            else:
                values_result = self.strict_fetch(item, check_for_errors=check_for_errors, flat_struct=True)

                fetched_length = len(values_result)
                if isinstance(item, slice):
                    item = slice(item.start, item.start + fetched_length, item.step)
                else:
                    item = slice(0, fetched_length)

                timestamps_result = timestamps_result_handle.fetch(
                    item, flat_struct=True, check_for_errors=check_for_errors
                )

                if timestamps_result is None:
                    raise QmNoResultsError("Failed to fetch timestamp results, please wait until results are ready")

                dtype = [
                    ("value", values_result.dtype),
                    ("timestamp", timestamps_result.dtype),
                ]
                combined = numpy.rec.fromarrays([values_result, timestamps_result], dtype=dtype)
                return cast(numpy.typing.NDArray[numpy.generic], combined.view(numpy.ndarray).astype(dtype))



import logging
import warnings

from qm.jobs.qm_job import QmJob  # noqa
from qm.type_hinting import DictQuaConfig
from qm.version import __version__  # noqa
from qm.logging_utils import config_loggers
from qm.jobs.job_queue_old_api import QmQueue
from qm.user_config import UserConfig  # noqa
from qm.jobs.pending_job import QmPendingJob  # noqa
from qm.quantum_machine import QuantumMachine  # noqa
from qm.program import Program, _Program, _ResultAnalysis  # noqa
from qm.api.models.compiler import CompilerOptionArguments  # noqa
from qm.quantum_machines_manager import QuantumMachinesManager  # noqa
from qm.serialization.generate_qua_script import generate_qua_script  # noqa
from qm.results import StreamingResultFetcher, SingleStreamingResultFetcher, MultipleStreamingResultFetcher  # noqa
from qm.simulate import (  # noqa
    InterOpxAddress,
    InterOpxChannel,
    InterOpxPairing,
    SimulationConfig,
    LoopbackInterface,
    ControllerConnection,
)

__all__ = [
    "QuantumMachinesManager",
    "QuantumMachine",
    "QmPendingJob",
    "QmJob",
    "StreamingResultFetcher",
    "SingleStreamingResultFetcher",
    "MultipleStreamingResultFetcher",
    "Program",
    "_ResultAnalysis",
    "CompilerOptionArguments",
    "generate_qua_script",
    "config_loggers",
    "UserConfig",
    "SimulationConfig",
    "InterOpxAddress",
    "InterOpxChannel",
    "ControllerConnection",
    "InterOpxPairing",
    "LoopbackInterface",
    "DictQuaConfig",
]

warnings.filterwarnings("default", category=DeprecationWarning, module="qm")
warnings.filterwarnings("ignore", category=DeprecationWarning, module="qm.grpc")

config = UserConfig.create_from_file()
config_loggers(config)


logger = logging.getLogger(__name__)
logger.info(f"Starting session: {config.SESSION_ID}")



from dataclasses import dataclass
from typing import Dict, List, Type, Tuple, AsyncIterator

from qm.api.v2.base_api_v2 import BaseApiV2
from qm.api.models.server_details import ConnectionDetails
from qm.StreamMetadata import StreamMetadata, StreamMetadataError, _get_stream_metadata_dict_from_proto_resp
from qm.grpc.v2 import (
    JobServiceStub,
    GetNamedResultRequest,
    GetJobResultStateRequest,
    GetJobResultSchemaRequest,
    GetProgramMetadataRequest,
    GetJobNamedResultHeaderRequest,
    GetNamedResultResponseGetNamedResultResponseSuccess,
    GetJobResultStateResponseGetJobResultStateResponseSuccess,
    GetJobResultSchemaResponseGetJobResultSchemaResponseSuccess,
    GetJobResultSchemaResponseGetJobResultSchemaResponseSuccessItem,
    GetJobNamedResultHeaderResponseGetJobNamedResultHeaderResponseSuccess,
)


@dataclass
class SchemaResponseItem:
    name: str
    simple_d_type: str
    is_single: bool
    expected_count: int
    shape: List[int]

    @classmethod
    def from_grpc(cls, data: GetJobResultSchemaResponseGetJobResultSchemaResponseSuccessItem) -> "SchemaResponseItem":
        return cls(
            name=data.name,
            simple_d_type=data.simple_dtype,
            is_single=data.is_single,
            expected_count=data.expected_count,
            shape=data.shape,
        )


@dataclass
class SchemaResponse:
    items: List[SchemaResponseItem]

    @classmethod
    def from_grpc(cls, data: GetJobResultSchemaResponseGetJobResultSchemaResponseSuccess) -> "SchemaResponse":
        return cls(items=[SchemaResponseItem.from_grpc(item) for item in data.items])


@dataclass
class NamedResultHeader:
    is_single: bool
    count_so_far: int
    simple_d_type: str
    has_dataloss: bool
    shape: List[int]
    has_execution_errors: bool = False

    @classmethod
    def from_grpc(
        cls, data: GetJobNamedResultHeaderResponseGetJobNamedResultHeaderResponseSuccess
    ) -> "NamedResultHeader":
        return cls(
            is_single=data.is_single,
            count_so_far=data.count_so_far,
            simple_d_type=data.simple_dtype,
            has_dataloss=data.has_data_loss,
            shape=data.shape,
        )


class JobResultApi(BaseApiV2[JobServiceStub]):
    def __init__(self, connection_details: ConnectionDetails, job_id: str):
        super().__init__(connection_details)
        self._id = job_id

    @property
    def _stub_class(self) -> Type[JobServiceStub]:
        return JobServiceStub

    @property
    def id(self) -> str:
        return self._id

    async def get_job_named_result(
        self, output_name: str, long_offset: int, limit: int
    ) -> AsyncIterator[GetNamedResultResponseGetNamedResultResponseSuccess]:
        request = GetNamedResultRequest(job_id=self._id, output_name=output_name, long_offset=long_offset, limit=limit)
        async for response in self._stub.get_named_result(request, timeout=self._timeout):
            yield response.success

    def get_named_header(self, output_name: str, flat_struct: bool) -> NamedResultHeader:
        request = GetJobNamedResultHeaderRequest(job_id=self._id, output_name=output_name, flat_format=flat_struct)
        response = self._run(self._stub.get_job_named_result_header(request, timeout=self._timeout))
        return NamedResultHeader.from_grpc(response)

    def get_program_metadata(self) -> Tuple[List[StreamMetadataError], Dict[str, StreamMetadata]]:
        request = GetProgramMetadataRequest(job_id=self._id)

        response = self._run(self._stub.get_program_metadata(request, timeout=self._timeout))
        metadata_errors = [
            StreamMetadataError(error.error, error.location)
            for error in response.program_stream_metadata.stream_metadata_extraction_error
        ]

        metadata_dict = _get_stream_metadata_dict_from_proto_resp(response.program_stream_metadata)
        return metadata_errors, metadata_dict

    def get_job_result_schema(self) -> SchemaResponse:
        request = GetJobResultSchemaRequest(job_id=self._id)
        response = self._run(self._stub.get_job_result_schema(request, timeout=self._timeout))
        return SchemaResponse.from_grpc(response)

    def get_job_state(self) -> GetJobResultStateResponseGetJobResultStateResponseSuccess:
        request = GetJobResultStateRequest(job_id=self._id)
        response = self._run(self._stub.get_job_result_state(request, timeout=self._timeout))
        return response



import sys
import logging

from qm.user_config import UserConfig


def config_loggers(user_config: UserConfig) -> None:
    package_logger = logging.getLogger("qm")
    if user_config.enable_user_stdout:
        handler = logging.StreamHandler(sys.stdout)
        handler.setFormatter(logging.Formatter(user_config.default_logging_format))
        package_logger.addHandler(handler)
        package_logger.setLevel(user_config.logging_level)
    if user_config.upload_logs and user_config.datadog_token:
        if not user_config.user_token:
            raise ValueError("No user token is defined")
        from qm.datadog_api import DatadogHandler

        package_logger.addHandler(
            DatadogHandler(
                user_id=user_config.user_token,
                organization=user_config.organization,
                user_token=user_config.datadog_token,
                session_id=str(user_config.SESSION_ID),
            )
        )


def set_logging_level(level: int) -> None:
    """Sets the logging level of the qm-qua module ("qm")
    See `Messages control <https://qm-docs.qualang.io/guides/error#messages-control>`__ for more information.

    Args:
        level: A string of either: 'DEBUG', 'INFO', 'WARNING', or
            'ERROR'. Can also accept standard python logging levels.
    """
    package_logger = logging.getLogger("qm")
    try:
        package_logger.setLevel(level)
    except (ValueError, TypeError):
        package_logger.warning(f"Failed to set log level. level '{level}' is not recognized")



import logging
from typing import List, Tuple, cast

from qm.simulate.interface import SimulatorInterface, SupportedConnectionTypes, get_opx_fem_number
from qm.grpc.frontend import (
    SimulationRequest,
    ExecutionRequestSimulateSimulationInterfaceNone,
    ExecutionRequestSimulateSimulationInterfaceLoopback,
    ExecutionRequestSimulateSimulationInterfaceLoopbackConnections,
)

logger = logging.getLogger(__name__)


class LoopbackInterface(SimulatorInterface[ExecutionRequestSimulateSimulationInterfaceLoopbackConnections]):
    """Creates a loopback interface for use in
    [qm.simulate.interface.SimulationConfig][].
    A loopback connects the output of the OPX into it's input. This can be defined
    directly using the ports or through the elements.

    Args:
        connections (list):
            List of tuples with loopback connections. Each tuple should represent physical connection between ports:

                    ``(from_controller: str, from_port: int, to_controller: str, to_port: int)``

        latency (int): The latency between the OPX outputs and its
            input.
        noisePower (float): How much noise to add to the input.

                ``(fromController: str, fromFEM: int, fromPort: int, toController: str, toFEM: int, toPort: int)``

            2. Virtual connection between elements:

                ``(fromQE: str, toQE: str, toQEInput: int)``
        latency: The latency between the OPX outputs and its input.
        noisePower: How much noise to add to the input.

    Example::

    >>> job = qmm.simulate(config, prog, SimulationConfig(
    >>>                   duration=20000,
    >>>                   # loopback from output 1 to input 2 of controller 1:
    >>>                   simulation_interface=LoopbackInterface([("con1", 1, "con1", 2)])
    """

    def __init__(self, connections: List[SupportedConnectionTypes], latency: int = 24, noisePower: float = 0.0):
        super().__init__(connections, noisePower)
        self._validate_latency(latency)
        self.latency = latency

    @property
    def connections(self) -> List[Tuple[str, int, int, str, int, int]]:
        connections = []
        for connection in self._connections:
            connections.append(
                (
                    connection.from_controller,
                    connection.from_fem,
                    connection.from_port,
                    connection.to_controller,
                    connection.to_fem,
                    connection.to_port,
                )
            )
        return connections

    @staticmethod
    def _validate_latency(latency: int) -> None:
        if (not isinstance(latency, int)) or latency < 0:
            raise Exception("latency must be a positive integer")

    @classmethod
    def _validate_and_standardize_single_connection(
        cls, connection: SupportedConnectionTypes
    ) -> ExecutionRequestSimulateSimulationInterfaceLoopbackConnections:
        if not connection:
            logger.warning("No loopback was defined, treating as no loopback.")
        if not isinstance(connection, tuple):
            raise Exception("each connection must be of type tuple")
        if len(connection) == 6:
            cls._validate_connection_type(
                connection,
                [str, int, int, str, int, int],
                "(from_controller, from_fem, from_port, to_controller, to_fem, to_port)",
            )
            tuple_6 = connection
            return ExecutionRequestSimulateSimulationInterfaceLoopbackConnections(
                from_controller=tuple_6[0],
                from_fem=tuple_6[1],
                from_port=tuple_6[2],
                to_controller=tuple_6[3],
                to_fem=tuple_6[4],
                to_port=tuple_6[5],
            )
        if len(connection) == 4:
            cls._validate_connection_type(
                connection, [str, int, str, int], "(from_controller, from_port, to_controller, to_port)"
            )
            tuple_4 = cast(Tuple[str, int, str, int], connection)
            return ExecutionRequestSimulateSimulationInterfaceLoopbackConnections(
                from_controller=tuple_4[0],
                from_fem=get_opx_fem_number(),
                from_port=tuple_4[1],
                to_controller=tuple_4[2],
                to_fem=get_opx_fem_number(),
                to_port=tuple_4[3],
            )
        if len(connection) == 3:
            cls._validate_connection_type(connection, [str, str, int], "(from_QE, to_QE, to_QEInput)")
            tuple_3 = cast(Tuple[str, str, int], connection)
            return ExecutionRequestSimulateSimulationInterfaceLoopbackConnections(
                from_controller=tuple_3[0],
                from_fem=-1,
                from_port=-1,
                to_controller=tuple_3[1],
                to_fem=-1,
                to_port=tuple_3[2],
            )
        raise Exception("connection should be tuple of length 3, 4 or 6")

    def update_simulate_request(
        self,
        request: SimulationRequest,
    ) -> SimulationRequest:
        if not self._connections:
            request.simulate.simulation_interface.none = ExecutionRequestSimulateSimulationInterfaceNone()
            return request

        request.simulate.simulation_interface.loopback = ExecutionRequestSimulateSimulationInterfaceLoopback(
            latency=self.latency, noise_power=self.noisePower, connections=self._connections
        )
        return request



from typing import Optional

from octave_sdk.octave import RFInput
from octave_sdk import IFMode, RFInputLOSource, RFInputRFSource


class ElementOutput:
    pass


class NoOutput(ElementOutput):
    pass


class DownconvertedOutput(ElementOutput):
    def __init__(self, client: RFInput):
        self._client = client

    def set_downconversion(
        self,
        lo_source: RFInputLOSource,
        lo_frequency: Optional[float] = None,
        if_mode_i: IFMode = IFMode.direct,
        if_mode_q: IFMode = IFMode.direct,
    ) -> None:
        """
        Sets the LO source for the downconverters.
        The LO source will be the one associated with the upconversion of element

        Args:
            lo_source:
            lo_frequency:
            if_mode_i:
            if_mode_q:
        """
        self._set_downconversion_lo(lo_source, lo_frequency)
        self._set_downconversion_if_mode(if_mode_i, if_mode_q)

    def _set_downconversion_lo(
        self,
        lo_source: RFInputLOSource,
        lo_frequency: Optional[float] = None,
    ) -> None:
        """
        Sets the LO source for the downconverters.
        If no value is given the LO source will be the one associated with the
        upconversion of element

        Args:
            lo_frequency:
            lo_source:
        """
        self._client.set_lo_source(lo_source, ignore_shared_errors=True)
        self._client.set_rf_source(RFInputRFSource.RF_in)
        is_internal = lo_source in {RFInputLOSource.Internal, RFInputLOSource.Analyzer}

        if lo_frequency is not None and is_internal:
            self._client.set_lo_frequency(source_name=lo_source, frequency=lo_frequency)

    def _set_downconversion_if_mode(self, if_mode_i: IFMode = IFMode.direct, if_mode_q: IFMode = IFMode.direct) -> None:
        self._client.set_if_mode_i(if_mode_i)
        self._client.set_if_mode_q(if_mode_q)



from typing import Dict, List, Tuple, Optional

from octave_sdk.octave import RFInput
from octave_sdk import Octave, OctaveOutput, OctaveLOSource

from qm.api.frontend_api import FrontendApi
from qm.elements.element_inputs import MixInputs
from qm.octave import CalibrationDB, QmOctaveConfig
from qm.api.models.capabilities import ServerCapabilities
from qm.utils.config_utils import get_fem_config, element_has_mix_inputs
from qm.octave.octave_manager import logger, get_device, get_loopbacks_from_pb
from qm.elements.up_converted_input import UpconvertedInput, UpconvertedInputNewApi
from qm.exceptions import OctaveCableSwapError, OctaveConnectionError, ElementUpconverterDeclarationError
from qm.grpc.qua_config import (
    QuaConfig,
    QuaConfigMatrix,
    QuaConfigElementDec,
    QuaConfigCorrectionEntry,
    QuaConfigGeneralPortReference,
)

OptionalOctaveInputPort = Optional[Tuple[str, int]]


class OctavesContainer:
    def __init__(self, pb_config: QuaConfig, octave_config: Optional[QmOctaveConfig] = None):
        self._pb_config = pb_config
        self._octave_config = octave_config or QmOctaveConfig()

        _qua_config_opx_to_octave_i = {}
        _qua_config_opx_to_octave_q = {}
        for octave_name, octave_qua_config in self._pb_config.v1_beta.octaves.items():
            for rf_idx, rf_config in octave_qua_config.rf_outputs.items():
                i_connection, q_connection = rf_config.i_connection, rf_config.q_connection
                _qua_config_opx_to_octave_i[(i_connection.controller, i_connection.fem, i_connection.number)] = (
                    octave_name,
                    rf_idx,
                )
                _qua_config_opx_to_octave_q[(q_connection.controller, q_connection.fem, q_connection.number)] = (
                    octave_name,
                    rf_idx,
                )
        self._qua_config_opx_to_octave_i = _qua_config_opx_to_octave_i
        self._qua_config_opx_to_octave_q = _qua_config_opx_to_octave_q

    def get_upconverter_port_ref(self, element_config: QuaConfigElementDec) -> OptionalOctaveInputPort:
        if element_config.rf_inputs:
            rf_input = list(element_config.rf_inputs.values())[0]
            return rf_input.device_name, rf_input.port
        if not element_has_mix_inputs(element_config):
            return None

        mix_inputs = element_config.mix_inputs
        element_i_port, element_q_port = mix_inputs.i, mix_inputs.q

        key_i = (element_i_port.controller, element_i_port.fem, element_i_port.number)
        key_q = (element_q_port.controller, element_q_port.fem, element_q_port.number)
        i_conn = self._qua_config_opx_to_octave_i.get(key_i)
        q_conn = self._qua_config_opx_to_octave_q.get(key_q)
        if i_conn is None and q_conn is None:
            if (self._qua_config_opx_to_octave_i.get(key_q) is not None) or (
                self._qua_config_opx_to_octave_q.get(key_i) is not None
            ):
                raise OctaveCableSwapError()

            return self._octave_config.get_octave_input_port(
                (element_i_port.controller, element_i_port.fem, element_i_port.number),
                (element_q_port.controller, element_q_port.fem, element_q_port.number),
            )
        if i_conn != q_conn:
            raise ElementUpconverterDeclarationError()
        return i_conn

    def _get_downconverter_client(self, outputs: Dict[str, QuaConfigGeneralPortReference]) -> Optional[RFInput]:
        if not outputs:
            return None
        for _, port_ref in outputs.items():
            if port_ref.device_name in self._pb_config.v1_beta.octaves:
                client = self._get_octave_client(port_ref.device_name)
                return client.rf_inputs[port_ref.port]
        raise OctaveConnectionError("No downconverter found for the given outputs.")

    def _get_loopbacks(self, octave_name: str) -> Dict[OctaveLOSource, OctaveOutput]:
        if octave_name in self._pb_config.v1_beta.octaves:
            pb_loopbacks = self._pb_config.v1_beta.octaves[octave_name].loopbacks
            return get_loopbacks_from_pb(pb_loopbacks, octave_name)
        return self._octave_config.get_lo_loopbacks_by_octave(octave_name)

    def create_mix_inputs(
        self,
        element_config: QuaConfigElementDec,
        name: str,
        frontend_api: FrontendApi,
        machine_id: str,
    ) -> MixInputs:
        port = self.get_upconverter_port_ref(element_config)
        if port is None:
            return MixInputs(name, element_config.mix_inputs, frontend_api, machine_id)

        octave_name, octave_port = port
        client = self._get_octave_client(octave_name)

        return UpconvertedInput(
            name,
            element_config.mix_inputs,
            frontend_api,
            machine_id,
            client=client.rf_outputs[octave_port],
            port=port,
            calibration_db=self._octave_config.calibration_db,
            gain=self._get_octave_gain(octave_name, octave_port),
            use_input_attenuators=self._get_octave_input_attenuators(octave_name, octave_port),
        )

    def create_new_api_upconverted_input(
        self,
        element_config: QuaConfigElementDec,
        name: str,
    ) -> Optional[UpconvertedInputNewApi]:
        port = self.get_upconverter_port_ref(element_config)
        if port is None:
            return None

        octave_name, octave_port = port
        client = self._get_octave_client(octave_name)
        gain = self._get_octave_gain(octave_name, octave_port)

        return UpconvertedInputNewApi(
            name,
            element_config.mix_inputs,
            client=client.rf_outputs[octave_port],
            port=port,
            calibration_db=self._octave_config.calibration_db,
            gain=gain,
            use_input_attenuators=self._get_octave_input_attenuators(octave_name, octave_port),
        )

    def _get_octave_gain(self, octave_name: str, port: int) -> Optional[float]:
        try:
            return self._pb_config.v1_beta.octaves[octave_name].rf_outputs[port].gain
        except KeyError:
            logger.warning(
                "No gain was specified. Setting gain to None. "
                "Please configure all the ports' gain values in the qua configuration"
            )
            return None

    def _get_octave_input_attenuators(self, octave_name: str, port: int) -> bool:
        try:
            return self._pb_config.v1_beta.octaves[octave_name].rf_outputs[port].input_attenuators
        except KeyError:
            return False

    def get_downconverter(self, outputs: Dict[str, QuaConfigGeneralPortReference]) -> Optional[RFInput]:
        return self._get_downconverter_client(outputs)

    def _get_octave_client(self, device_name: str) -> Octave:
        device_connection_info = self._octave_config.devices[device_name]
        loopbacks = self._get_loopbacks(device_name)
        return get_device(
            device_connection_info,
            loop_backs=loopbacks,
            octave_name=device_name,
            fan=self._octave_config.fan,
        )


def load_config_from_calibration_db(
    pb_config: QuaConfig, calibration_db: CalibrationDB, octave_config: QmOctaveConfig, capabilities: ServerCapabilities
) -> QuaConfig:
    octaves_container = OctavesContainer(pb_config, octave_config)

    logger.debug("Loading mixer calibration data onto the config")

    for element_name, element in pb_config.v1_beta.elements.items():
        if not element_has_mix_inputs(element):
            continue

        mix_inputs = element.mix_inputs
        lo_freq = mix_inputs.lo_frequency or mix_inputs.lo_frequency_double
        if not lo_freq:
            logger.debug(f"Element '{element_name}' has no LO frequency specified")
            continue

        octave_channel = octaves_container.get_upconverter_port_ref(element)
        if octave_channel is None:
            logger.debug(f"Element '{element_name}' is not connected to Octave")
            continue

        try:
            output_gain = pb_config.v1_beta.octaves[octave_channel[0]].rf_outputs[octave_channel[1]].gain
        except KeyError:
            logger.warning(
                "No gain was specified. Setting gain to None. "
                "Please configure all the ports' gain values in the qua configuration"
            )
            output_gain = None

        lo_cal = calibration_db.get_lo_cal(octave_channel, lo_freq, output_gain)
        if lo_cal is None:
            logger.debug(
                f"the calibration db has no LO cal for element '{element_name}' (lo_freq = {lo_freq / 1e9:0.3f} GHz)"
            )
            continue

        i_port, q_port = mix_inputs.i, mix_inputs.q
        i_controller_config = get_fem_config(pb_config, i_port)
        i_controller_config.analog_outputs[i_port.number].offset = lo_cal.get_i0()
        q_controller_config = get_fem_config(pb_config, q_port)
        q_controller_config.analog_outputs[q_port.number].offset = lo_cal.get_q0()

        if i_port.controller in pb_config.v1_beta.controllers:
            pb_config.v1_beta.controllers[i_port.controller].analog_outputs[i_port.number].offset = lo_cal.get_i0()
        if q_port.controller in pb_config.v1_beta.controllers:
            pb_config.v1_beta.controllers[q_port.controller].analog_outputs[q_port.number].offset = lo_cal.get_q0()

        # Now we go over all the IF frequencies we find and set them. Not sure
        # when an IF frequency different from the element's 'intermediate_frequency'
        # will be used. Maybe when static frequency change happens, the gateway takes
        # the correction from this updated config.
        if_calibrations_for_curr_lo = calibration_db.get_all_if_cal_for_lo(octave_channel, lo_freq, output_gain)

        # We are expected to put all these calibrations in the element's mixer
        curr_mixer = mix_inputs.mixer

        if curr_mixer not in pb_config.v1_beta.mixers:
            logger.debug(f"Element '{element_name}' is using mixer '{curr_mixer}' which is not found.")
            continue

        old_if_cals = pb_config.v1_beta.mixers[curr_mixer]
        new_if_cals: List[QuaConfigCorrectionEntry] = []
        frequency_idx = int(capabilities.supports_double_frequency)  # This is to shorten many if-else in the code using
        # a trinary expression
        for if_freq, if_cal in if_calibrations_for_curr_lo.items():
            curr_new_calibration = QuaConfigCorrectionEntry(
                frequency=int(abs(if_freq)),
                frequency_double=[0.0, float(abs(if_freq))][frequency_idx],
                lo_frequency=int(lo_freq),
                lo_frequency_double=[0.0, float(lo_freq)][frequency_idx],
                correction=QuaConfigMatrix(*if_cal.get_correction()),
                frequency_negative=if_freq < 0,
            )
            new_if_cals.append(curr_new_calibration)

        for old_if_cal in old_if_cals.correction:
            if lo_freq not in {old_if_cal.lo_frequency, old_if_cal.lo_frequency_double}:
                continue

            sign = (-1) ** old_if_cal.frequency_negative
            old_if_freq = (old_if_cal.frequency or old_if_cal.frequency_double) * sign
            if old_if_freq in if_calibrations_for_curr_lo:
                continue

            new_if_cals.append(old_if_cal)
            logger.debug(
                f"Could not find calibration value for LO frequency {lo_freq} and intermediate_frequency {old_if_freq}"
            )

        pb_config.v1_beta.mixers[curr_mixer].correction = new_if_cals
    return pb_config
